from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx.domains.c._ast import (
    ASTAlignofExpr,
    ASTArray,
    ASTAssignmentExpr,
    ASTBinOpExpr,
    ASTBooleanLiteral,
    ASTBracedInitList,
    ASTCastExpr,
    ASTCharLiteral,
    ASTDeclaration,
    ASTDeclaratorNameBitField,
    ASTDeclaratorNameParam,
    ASTDeclaratorParen,
    ASTDeclaratorPtr,
    ASTDeclSpecs,
    ASTDeclSpecsSimple,
    ASTEnum,
    ASTEnumerator,
    ASTFallbackExpr,
    ASTFunctionParameter,
    ASTIdentifier,
    ASTIdExpression,
    ASTInitializer,
    ASTMacro,
    ASTMacroParameter,
    ASTNestedName,
    ASTNumberLiteral,
    ASTParameters,
    ASTParenExpr,
    ASTParenExprList,
    ASTPostfixArray,
    ASTPostfixCallExpr,
    ASTPostfixDec,
    ASTPostfixExpr,
    ASTPostfixInc,
    ASTPostfixMemberOfPointer,
    ASTSizeofExpr,
    ASTSizeofType,
    ASTStringLiteral,
    ASTStruct,
    ASTTrailingTypeSpecFundamental,
    ASTTrailingTypeSpecName,
    ASTType,
    ASTTypeWithInit,
    ASTUnaryOpExpr,
    ASTUnion,
)
from sphinx.domains.c._ids import (
    _expression_assignment_ops,
    _expression_bin_ops,
    _expression_unary_ops,
    _keywords,
    _simple_type_specifiers_re,
    _string_re,
)
from sphinx.util.cfamily import (
    ASTAttributeList,
    BaseParser,
    DefinitionError,
    UnsupportedMultiCharacterCharLiteral,
    binary_literal_re,
    char_literal_re,
    float_literal_re,
    float_literal_suffix_re,
    hex_literal_re,
    identifier_re,
    integer_literal_re,
    integers_literal_suffix_re,
    octal_literal_re,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any

    from sphinx.domains.c._ast import (
        ASTDeclarator,
        ASTExpression,
        ASTLiteral,
        ASTPostfixOp,
        ASTTrailingTypeSpec,
        DeclarationType,
    )


class DefinitionParser(BaseParser):
    @property
    def language(self) -> str:
        return 'C'

    @property
    def id_attributes(self) -> Sequence[str]:
        return self.config.c_id_attributes

    @property
    def paren_attributes(self) -> Sequence[str]:
        return self.config.c_paren_attributes

    def _parse_string(self) -> str | None:
        if self.current_char != '"':
            return None
        start_pos = self.pos
        self.pos += 1
        escape = False
        while True:
            if self.eof:
                self.fail('Unexpected end during inside string.')
            elif self.current_char == '"' and not escape:
                self.pos += 1
                break
            elif self.current_char == '\\':
                escape = True
            else:
                escape = False
            self.pos += 1
        return self.definition[start_pos : self.pos]

    def _parse_literal(self) -> ASTLiteral | None:
        # -> integer-literal
        #  | character-literal
        #  | floating-literal
        #  | string-literal
        #  | boolean-literal -> "false" | "true"
        self.skip_ws()
        if self.skip_word('true'):
            return ASTBooleanLiteral(True)
        if self.skip_word('false'):
            return ASTBooleanLiteral(False)
        pos = self.pos
        if self.match(float_literal_re):
            self.match(float_literal_suffix_re)
            return ASTNumberLiteral(self.definition[pos : self.pos])
        for regex in (
            binary_literal_re,
            hex_literal_re,
            integer_literal_re,
            octal_literal_re,
        ):
            if self.match(regex):
                self.match(integers_literal_suffix_re)
                return ASTNumberLiteral(self.definition[pos : self.pos])

        string = self._parse_string()
        if string is not None:
            return ASTStringLiteral(string)

        # character-literal
        if self.match(char_literal_re):
            prefix = self.last_match.group(1)  # may be None when no prefix
            data = self.last_match.group(2)
            try:
                return ASTCharLiteral(prefix, data)
            except UnicodeDecodeError as e:
                self.fail(
                    'Can not handle character literal. Internal error was: %s' % e
                )
            except UnsupportedMultiCharacterCharLiteral:
                self.fail(
                    'Can not handle character literal'
                    ' resulting in multiple decoded characters.'
                )
        return None

    def _parse_paren_expression(self) -> ASTExpression | None:
        # "(" expression ")"
        if self.current_char != '(':
            return None
        self.pos += 1
        res = self._parse_expression()
        self.skip_ws()
        if not self.skip_string(')'):
            self.fail("Expected ')' in end of parenthesized expression.")
        return ASTParenExpr(res)

    def _parse_primary_expression(self) -> ASTExpression | None:
        # literal
        # "(" expression ")"
        # id-expression -> we parse this with _parse_nested_name
        self.skip_ws()
        res: ASTExpression | None = self._parse_literal()
        if res is not None:
            return res
        res = self._parse_paren_expression()
        if res is not None:
            return res
        nn = self._parse_nested_name()
        if nn is not None:
            return ASTIdExpression(nn)
        return None

    def _parse_initializer_list(
        self, name: str, open: str, close: str
    ) -> tuple[list[ASTExpression] | None, bool | None]:
        # Parse open and close with the actual initializer-list in between
        # -> initializer-clause '...'[opt]
        #  | initializer-list ',' initializer-clause '...'[opt]
        # TODO: designators
        self.skip_ws()
        if not self.skip_string_and_ws(open):
            return None, None
        if self.skip_string(close):
            return [], False

        exprs = []
        trailing_comma = False
        while True:
            self.skip_ws()
            expr = self._parse_expression()
            self.skip_ws()
            exprs.append(expr)
            self.skip_ws()
            if self.skip_string(close):
                break
            if not self.skip_string_and_ws(','):
                self.fail(f"Error in {name}, expected ',' or '{close}'.")
            if self.current_char == close == '}':
                self.pos += 1
                trailing_comma = True
                break
        return exprs, trailing_comma

    def _parse_paren_expression_list(self) -> ASTParenExprList | None:
        # -> '(' expression-list ')'
        # though, we relax it to also allow empty parens
        # as it's needed in some cases
        #
        # expression-list
        # -> initializer-list
        exprs, _trailing_comma = self._parse_initializer_list(
            'parenthesized expression-list', '(', ')'
        )
        if exprs is None:
            return None
        return ASTParenExprList(exprs)

    def _parse_braced_init_list(self) -> ASTBracedInitList | None:
        # -> '{' initializer-list ','[opt] '}'
        #  | '{' '}'
        exprs, trailing_comma = self._parse_initializer_list(
            'braced-init-list', '{', '}'
        )
        if exprs is None:
            return None
        return ASTBracedInitList(exprs, trailing_comma)

    def _parse_postfix_expression(self) -> ASTPostfixExpr:
        # -> primary
        #  | postfix "[" expression "]"
        #  | postfix "[" braced-init-list [opt] "]"
        #  | postfix "(" expression-list [opt] ")"
        #  | postfix "." id-expression  // taken care of in primary by nested name
        #  | postfix "->" id-expression
        #  | postfix "++"
        #  | postfix "--"

        prefix = self._parse_primary_expression()

        # and now parse postfixes
        post_fixes: list[ASTPostfixOp] = []
        while True:
            self.skip_ws()
            if self.skip_string_and_ws('['):
                expr = self._parse_expression()
                self.skip_ws()
                if not self.skip_string(']'):
                    self.fail("Expected ']' in end of postfix expression.")
                post_fixes.append(ASTPostfixArray(expr))
                continue
            if self.skip_string('->'):
                if self.skip_string('*'):
                    # don't steal the arrow
                    self.pos -= 3
                else:
                    name = self._parse_nested_name()
                    post_fixes.append(ASTPostfixMemberOfPointer(name))
                    continue
            if self.skip_string('++'):
                post_fixes.append(ASTPostfixInc())
                continue
            if self.skip_string('--'):
                post_fixes.append(ASTPostfixDec())
                continue
            lst = self._parse_paren_expression_list()
            if lst is not None:
                post_fixes.append(ASTPostfixCallExpr(lst))
                continue
            break
        return ASTPostfixExpr(prefix, post_fixes)

    def _parse_unary_expression(self) -> ASTExpression:
        # -> postfix
        #  | "++" cast
        #  | "--" cast
        #  | unary-operator cast -> (* | & | + | - | ! | ~) cast
        # The rest:
        #  | "sizeof" unary
        #  | "sizeof" "(" type-id ")"
        #  | "alignof" "(" type-id ")"
        self.skip_ws()
        for op in _expression_unary_ops:
            # TODO: hmm, should we be able to backtrack here?
            if op[0] in 'cn':
                res = self.skip_word(op)
            else:
                res = self.skip_string(op)
            if res:
                expr = self._parse_cast_expression()
                return ASTUnaryOpExpr(op, expr)
        if self.skip_word_and_ws('sizeof'):
            if self.skip_string_and_ws('('):
                typ = self._parse_type(named=False)
                self.skip_ws()
                if not self.skip_string(')'):
                    self.fail("Expecting ')' to end 'sizeof'.")
                return ASTSizeofType(typ)
            expr = self._parse_unary_expression()
            return ASTSizeofExpr(expr)
        if self.skip_word_and_ws('alignof'):
            if not self.skip_string_and_ws('('):
                self.fail("Expecting '(' after 'alignof'.")
            typ = self._parse_type(named=False)
            self.skip_ws()
            if not self.skip_string(')'):
                self.fail("Expecting ')' to end 'alignof'.")
            return ASTAlignofExpr(typ)
        return self._parse_postfix_expression()

    def _parse_cast_expression(self) -> ASTExpression:
        # -> unary  | "(" type-id ")" cast
        pos = self.pos
        self.skip_ws()
        if self.skip_string('('):
            try:
                typ = self._parse_type(False)
                if not self.skip_string(')'):
                    self.fail("Expected ')' in cast expression.")
                expr = self._parse_cast_expression()
                return ASTCastExpr(typ, expr)
            except DefinitionError as ex_cast:
                self.pos = pos
                try:
                    return self._parse_unary_expression()
                except DefinitionError as ex_unary:
                    errs = [
                        (ex_cast, 'If type cast expression'),
                        (ex_unary, 'If unary expression'),
                    ]
                    raise self._make_multi_error(
                        errs, 'Error in cast expression.'
                    ) from ex_unary
        else:
            return self._parse_unary_expression()

    def _parse_logical_or_expression(self) -> ASTExpression:
        # logical-or     = logical-and      ||
        # logical-and    = inclusive-or     &&
        # inclusive-or   = exclusive-or     |
        # exclusive-or   = and              ^
        # and            = equality         &
        # equality       = relational       ==, !=
        # relational     = shift            <, >, <=, >=
        # shift          = additive         <<, >>
        # additive       = multiplicative   +, -
        # multiplicative = pm               *, /, %
        # pm             = cast             .*, ->*
        def _parse_bin_op_expr(self: DefinitionParser, op_id: int) -> ASTExpression:
            if op_id + 1 == len(_expression_bin_ops):
                parser = self._parse_cast_expression
            else:

                def parser() -> ASTExpression:
                    return _parse_bin_op_expr(self, op_id + 1)

            exprs = []
            ops = []
            exprs.append(parser())
            while True:
                self.skip_ws()
                pos = self.pos
                one_more = False
                for op in _expression_bin_ops[op_id]:
                    if op[0] in 'abcnox':
                        if not self.skip_word(op):
                            continue
                    else:
                        if not self.skip_string(op):
                            continue
                    if op == self.current_char == '&':
                        # don't split the && 'token'
                        self.pos -= 1
                        # and btw. && has lower precedence, so we are done
                        break
                    try:
                        expr = parser()
                        exprs.append(expr)
                        ops.append(op)
                        one_more = True
                        break
                    except DefinitionError:
                        self.pos = pos
                if not one_more:
                    break
            return ASTBinOpExpr(exprs, ops)  # type: ignore[return-value]

        return _parse_bin_op_expr(self, 0)

    def _parse_conditional_expression_tail(
        self, or_expr_head: Any
    ) -> ASTExpression | None:
        # -> "?" expression ":" assignment-expression
        return None

    def _parse_assignment_expression(self) -> ASTExpression:
        # -> conditional-expression
        #  | logical-or-expression assignment-operator initializer-clause
        # -> conditional-expression ->
        #     logical-or-expression
        #   | logical-or-expression "?" expression ":" assignment-expression
        #   | logical-or-expression assignment-operator initializer-clause
        exprs = []
        ops = []
        or_expr = self._parse_logical_or_expression()
        exprs.append(or_expr)
        # TODO: handle ternary with _parse_conditional_expression_tail
        while True:
            one_more = False
            self.skip_ws()
            for op in _expression_assignment_ops:
                if op[0] in 'abcnox':
                    if not self.skip_word(op):
                        continue
                else:
                    if not self.skip_string(op):
                        continue
                expr = self._parse_logical_or_expression()
                exprs.append(expr)
                ops.append(op)
                one_more = True
            if not one_more:
                break
        return ASTAssignmentExpr(exprs, ops)

    def _parse_constant_expression(self) -> ASTExpression:
        # -> conditional-expression
        or_expr = self._parse_logical_or_expression()
        # TODO: use _parse_conditional_expression_tail
        return or_expr

    def _parse_expression(self) -> ASTExpression:
        # -> assignment-expression
        #  | expression "," assignment-expression
        # TODO: actually parse the second production
        return self._parse_assignment_expression()

    def _parse_expression_fallback(
        self, end: list[str], parser: Callable[[], ASTExpression], allow: bool = True
    ) -> ASTExpression:
        # Stupidly "parse" an expression.
        # 'end' should be a list of characters which ends the expression.

        # first try to use the provided parser
        prev_pos = self.pos
        try:
            return parser()
        except DefinitionError as e:
            # some places (e.g., template parameters) we really don't want to use fallback,
            # and for testing we may want to globally disable it
            if not allow or not self.allowFallbackExpressionParsing:
                raise
            self.warn(
                'Parsing of expression failed. Using fallback parser.'
                ' Error was:\n%s' % e
            )
            self.pos = prev_pos
        # and then the fallback scanning
        assert end is not None
        self.skip_ws()
        start_pos = self.pos
        if self.match(_string_re):
            value = self.matched_text
        else:
            # TODO: add handling of more bracket-like things, and quote handling
            brackets = {'(': ')', '{': '}', '[': ']'}
            symbols: list[str] = []
            while not self.eof:
                if len(symbols) == 0 and self.current_char in end:
                    break
                if self.current_char in brackets:
                    symbols.append(brackets[self.current_char])
                elif len(symbols) > 0 and self.current_char == symbols[-1]:
                    symbols.pop()
                self.pos += 1
            if len(end) > 0 and self.eof:
                self.fail(
                    'Could not find end of expression starting at %d.' % start_pos
                )
            value = self.definition[start_pos : self.pos].strip()
        return ASTFallbackExpr(value.strip())

    def _parse_nested_name(self) -> ASTNestedName:
        names: list[Any] = []

        self.skip_ws()
        rooted = False
        if self.skip_string('.'):
            rooted = True
        while 1:
            self.skip_ws()
            if not self.match(identifier_re):
                self.fail('Expected identifier in nested name.')
            identifier = self.matched_text
            # make sure there isn't a keyword
            if identifier in _keywords:
                self.fail(
                    'Expected identifier in nested name, got keyword: %s' % identifier
                )
            if self.matched_text in self.config.c_extra_keywords:
                msg = (
                    'Expected identifier, got user-defined keyword: %s.'
                    ' Remove it from c_extra_keywords to allow it as identifier.\n'
                    'Currently c_extra_keywords is %s.'
                )
                self.fail(msg % (self.matched_text, str(self.config.c_extra_keywords)))
            ident = ASTIdentifier(identifier)
            names.append(ident)

            self.skip_ws()
            if not self.skip_string('.'):
                break
        return ASTNestedName(names, rooted)

    def _parse_simple_type_specifier(self) -> str | None:
        if self.match(_simple_type_specifiers_re):
            return self.matched_text
        for t in ('bool', 'complex', 'imaginary'):
            if t in self.config.c_extra_keywords:
                if self.skip_word(t):
                    return t
        return None

    def _parse_simple_type_specifiers(self) -> ASTTrailingTypeSpecFundamental | None:
        names: list[str] = []

        self.skip_ws()
        while True:
            t = self._parse_simple_type_specifier()
            if t is None:
                break
            names.append(t)
            self.skip_ws()
        if len(names) == 0:
            return None
        return ASTTrailingTypeSpecFundamental(names)

    def _parse_trailing_type_spec(self) -> ASTTrailingTypeSpec:
        # fundamental types, https://en.cppreference.com/w/c/language/type
        # and extensions
        self.skip_ws()
        res = self._parse_simple_type_specifiers()
        if res is not None:
            return res

        # prefixed
        prefix = None
        self.skip_ws()
        for k in ('struct', 'enum', 'union'):
            if self.skip_word_and_ws(k):
                prefix = k
                break

        nested_name = self._parse_nested_name()
        return ASTTrailingTypeSpecName(prefix, nested_name)

    def _parse_parameters(self, param_mode: str) -> ASTParameters | None:
        self.skip_ws()
        if not self.skip_string('('):
            if param_mode == 'function':
                self.fail('Expecting "(" in parameters.')
            else:
                return None

        args = []
        self.skip_ws()
        if not self.skip_string(')'):
            while 1:
                self.skip_ws()
                if self.skip_string('...'):
                    args.append(ASTFunctionParameter(None, True))
                    self.skip_ws()
                    if not self.skip_string(')'):
                        self.fail('Expected ")" after "..." in parameters.')
                    break
                # note: it seems that function arguments can always be named,
                # even in function pointers and similar.
                arg = self._parse_type_with_init(outer=None, named='single')
                # TODO: parse default parameters # TODO: didn't we just do that?
                args.append(ASTFunctionParameter(arg))

                self.skip_ws()
                if self.skip_string(','):
                    continue
                if self.skip_string(')'):
                    break
                self.fail(
                    f'Expecting "," or ")" in parameters, got "{self.current_char}".'
                )

        attrs = self._parse_attribute_list()
        return ASTParameters(args, attrs)

    def _parse_decl_specs_simple(
        self, outer: str | None, typed: bool
    ) -> ASTDeclSpecsSimple:
        """Just parse the simple ones."""
        storage = None
        thread_local = None
        inline = None
        restrict = None
        volatile = None
        const = None
        attrs = []
        while 1:  # accept any permutation of a subset of some decl-specs
            self.skip_ws()
            if not storage:
                if outer == 'member':
                    if self.skip_word('auto'):
                        storage = 'auto'
                        continue
                    if self.skip_word('register'):
                        storage = 'register'
                        continue
                if outer in {'member', 'function'}:
                    if self.skip_word('static'):
                        storage = 'static'
                        continue
                    if self.skip_word('extern'):
                        storage = 'extern'
                        continue
            if outer == 'member' and not thread_local:
                if self.skip_word('thread_local'):
                    thread_local = 'thread_local'
                    continue
                if self.skip_word('_Thread_local'):
                    thread_local = '_Thread_local'
                    continue
            if outer == 'function' and not inline:
                inline = self.skip_word('inline')
                if inline:
                    continue

            if not restrict and typed:
                restrict = self.skip_word('restrict')
                if restrict:
                    continue
            if not volatile and typed:
                volatile = self.skip_word('volatile')
                if volatile:
                    continue
            if not const and typed:
                const = self.skip_word('const')
                if const:
                    continue
            attr = self._parse_attribute()
            if attr:
                attrs.append(attr)
                continue
            break
        return ASTDeclSpecsSimple(
            storage,
            thread_local,
            inline,
            restrict,
            volatile,
            const,
            ASTAttributeList(attrs),
        )

    def _parse_decl_specs(self, outer: str | None, typed: bool = True) -> ASTDeclSpecs:
        if outer:
            if outer not in {'type', 'member', 'function'}:
                raise Exception('Internal error, unknown outer "%s".' % outer)
        left_specs = self._parse_decl_specs_simple(outer, typed)
        right_specs = None

        if typed:
            trailing = self._parse_trailing_type_spec()
            right_specs = self._parse_decl_specs_simple(outer, typed)
        else:
            trailing = None
        return ASTDeclSpecs(outer, left_specs, right_specs, trailing)

    def _parse_declarator_name_suffix(
        self, named: bool | str, param_mode: str, typed: bool
    ) -> ASTDeclarator:
        assert named in {True, False, 'single'}
        # now we should parse the name, and then suffixes
        if named == 'single':
            if self.match(identifier_re):
                if self.matched_text in _keywords:
                    self.fail(
                        'Expected identifier, got keyword: %s' % self.matched_text
                    )
                if self.matched_text in self.config.c_extra_keywords:
                    msg = (
                        'Expected identifier, got user-defined keyword: %s. '
                        'Remove it from c_extra_keywords to allow it as identifier.\n'
                        'Currently c_extra_keywords is %s.'
                    )
                    self.fail(
                        msg % (self.matched_text, str(self.config.c_extra_keywords))
                    )
                identifier = ASTIdentifier(self.matched_text)
                decl_id = ASTNestedName([identifier], rooted=False)
            else:
                decl_id = None
        elif named:
            decl_id = self._parse_nested_name()
        else:
            decl_id = None
        array_ops = []
        while 1:
            self.skip_ws()
            if typed and self.skip_string('['):
                self.skip_ws()
                static = False
                const = False
                volatile = False
                restrict = False
                while True:
                    if not static:
                        if self.skip_word_and_ws('static'):
                            static = True
                            continue
                    if not const:
                        if self.skip_word_and_ws('const'):
                            const = True
                            continue
                    if not volatile:
                        if self.skip_word_and_ws('volatile'):
                            volatile = True
                            continue
                    if not restrict:
                        if self.skip_word_and_ws('restrict'):
                            restrict = True
                            continue
                    break
                vla = False if static else self.skip_string_and_ws('*')
                if vla:
                    if not self.skip_string(']'):
                        self.fail("Expected ']' in end of array operator.")
                    size = None
                else:
                    if self.skip_string(']'):
                        size = None
                    else:
                        parser = self._parse_expression
                        size = self._parse_expression_fallback([']'], parser)
                        self.skip_ws()
                        if not self.skip_string(']'):
                            self.fail("Expected ']' in end of array operator.")
                array_ops.append(ASTArray(static, const, volatile, restrict, vla, size))
            else:
                break
        param = self._parse_parameters(param_mode)
        if param is None and len(array_ops) == 0:
            # perhaps a bit-field
            if named and param_mode == 'type' and typed:
                self.skip_ws()
                if self.skip_string(':'):
                    size = self._parse_constant_expression()
                    return ASTDeclaratorNameBitField(declId=decl_id, size=size)
        return ASTDeclaratorNameParam(declId=decl_id, arrayOps=array_ops, param=param)

    def _parse_declarator(
        self, named: bool | str, param_mode: str, typed: bool = True
    ) -> ASTDeclarator:
        # 'typed' here means 'parse return type stuff'
        if param_mode not in {'type', 'function'}:
            raise Exception("Internal error, unknown param_mode '%s'." % param_mode)
        prev_errors = []
        self.skip_ws()
        if typed and self.skip_string('*'):
            self.skip_ws()
            restrict = False
            volatile = False
            const = False
            attrs = []
            while 1:
                if not restrict:
                    restrict = self.skip_word_and_ws('restrict')
                    if restrict:
                        continue
                if not volatile:
                    volatile = self.skip_word_and_ws('volatile')
                    if volatile:
                        continue
                if not const:
                    const = self.skip_word_and_ws('const')
                    if const:
                        continue
                attr = self._parse_attribute()
                if attr is not None:
                    attrs.append(attr)
                    continue
                break
            next = self._parse_declarator(named, param_mode, typed)
            return ASTDeclaratorPtr(
                next=next,
                restrict=restrict,
                volatile=volatile,
                const=const,
                attrs=ASTAttributeList(attrs),
            )
        if typed and self.current_char == '(':  # note: peeking, not skipping
            # maybe this is the beginning of params, try that first,
            # otherwise assume it's noptr->declarator > ( ptr-declarator )
            pos = self.pos
            try:
                # assume this is params
                res = self._parse_declarator_name_suffix(named, param_mode, typed)
                return res
            except DefinitionError as ex_param_qual:
                msg = 'If declarator-id with parameters'
                if param_mode == 'function':
                    msg += " (e.g., 'void f(int arg)')"
                prev_errors.append((ex_param_qual, msg))
                self.pos = pos
                try:
                    assert self.current_char == '('
                    self.skip_string('(')
                    # TODO: hmm, if there is a name, it must be in inner, right?
                    # TODO: hmm, if there must be parameters, they must b
                    # inside, right?
                    inner = self._parse_declarator(named, param_mode, typed)
                    if not self.skip_string(')'):
                        self.fail('Expected \')\' in "( ptr-declarator )"')
                    next = self._parse_declarator(
                        named=False, param_mode='type', typed=typed
                    )
                    return ASTDeclaratorParen(inner=inner, next=next)
                except DefinitionError as ex_no_ptr_paren:
                    self.pos = pos
                    msg = 'If parenthesis in noptr-declarator'
                    if param_mode == 'function':
                        msg += " (e.g., 'void (*f(int arg))(double)')"
                    prev_errors.append((ex_no_ptr_paren, msg))
                    header = 'Error in declarator'
                    raise self._make_multi_error(
                        prev_errors, header
                    ) from ex_no_ptr_paren
        pos = self.pos
        try:
            return self._parse_declarator_name_suffix(named, param_mode, typed)
        except DefinitionError as e:
            self.pos = pos
            prev_errors.append((e, 'If declarator-id'))
            header = 'Error in declarator or parameters'
            raise self._make_multi_error(prev_errors, header) from e

    def _parse_initializer(
        self, outer: str | None = None, allow_fallback: bool = True
    ) -> ASTInitializer | None:
        self.skip_ws()
        if outer == 'member' and False:  # NoQA: SIM223,TD005  # TODO
            braced_init = self._parse_braced_init_list()
            if braced_init is not None:
                return ASTInitializer(braced_init, hasAssign=False)

        if not self.skip_string('='):
            return None

        braced_init = self._parse_braced_init_list()
        if braced_init is not None:
            return ASTInitializer(braced_init)

        if outer == 'member':
            fallback_end: list[str] = []
        elif outer is None:  # function parameter
            fallback_end = [',', ')']
        else:
            self.fail(
                "Internal error, initializer for outer '%s' not implemented." % outer
            )

        def parser() -> ASTExpression:
            return self._parse_assignment_expression()

        value = self._parse_expression_fallback(
            fallback_end, parser, allow=allow_fallback
        )
        return ASTInitializer(value)

    def _parse_type(self, named: bool | str, outer: str | None = None) -> ASTType:
        """named=False|'single'|True: 'single' is e.g., for function objects which
        doesn't need to name the arguments, but otherwise is a single name
        """
        if outer:  # always named
            if outer not in {'type', 'member', 'function'}:
                raise Exception('Internal error, unknown outer "%s".' % outer)
            assert named

        if outer == 'type':
            # We allow type objects to just be a name.
            prev_errors = []
            start_pos = self.pos
            # first try without the type
            try:
                decl_specs = self._parse_decl_specs(outer=outer, typed=False)
                decl = self._parse_declarator(named=True, param_mode=outer, typed=False)
                self.assert_end(allowSemicolon=True)
            except DefinitionError as ex_untyped:
                desc = 'If just a name'
                prev_errors.append((ex_untyped, desc))
                self.pos = start_pos
                try:
                    decl_specs = self._parse_decl_specs(outer=outer)
                    decl = self._parse_declarator(named=True, param_mode=outer)
                except DefinitionError as ex_typed:
                    self.pos = start_pos
                    desc = 'If typedef-like declaration'
                    prev_errors.append((ex_typed, desc))
                    # Retain the else branch for easier debugging.
                    # TODO: it would be nice to save the previous stacktrace
                    #       and output it here.
                    if True:
                        header = 'Type must be either just a name or a '
                        header += 'typedef-like declaration.'
                        raise self._make_multi_error(prev_errors, header) from ex_typed
                    else:
                        # For testing purposes.
                        # do it again to get the proper traceback (how do you
                        # reliably save a traceback when an exception is
                        # constructed?)
                        self.pos = start_pos
                        typed = True
                        decl_specs = self._parse_decl_specs(outer=outer, typed=typed)
                        decl = self._parse_declarator(
                            named=True, param_mode=outer, typed=typed
                        )
        elif outer == 'function':
            decl_specs = self._parse_decl_specs(outer=outer)
            decl = self._parse_declarator(named=True, param_mode=outer)
        else:
            param_mode = 'type'
            if outer == 'member':  # i.e., member
                named = True
            decl_specs = self._parse_decl_specs(outer=outer)
            decl = self._parse_declarator(named=named, param_mode=param_mode)
        return ASTType(decl_specs, decl)

    def _parse_type_with_init(
        self, named: bool | str, outer: str | None
    ) -> ASTTypeWithInit:
        if outer:
            assert outer in {'type', 'member', 'function'}
        type = self._parse_type(outer=outer, named=named)
        init = self._parse_initializer(outer=outer)
        return ASTTypeWithInit(type, init)

    def _parse_macro(self) -> ASTMacro:
        self.skip_ws()
        ident = self._parse_nested_name()
        if ident is None:
            self.fail('Expected identifier in macro definition.')
        self.skip_ws()
        if not self.skip_string_and_ws('('):
            return ASTMacro(ident, None)
        if self.skip_string(')'):
            return ASTMacro(ident, [])
        args = []
        while 1:
            self.skip_ws()
            if self.skip_string('...'):
                args.append(ASTMacroParameter(None, True))
                self.skip_ws()
                if not self.skip_string(')'):
                    self.fail('Expected ")" after "..." in macro parameters.')
                break
            if not self.match(identifier_re):
                self.fail('Expected identifier in macro parameters.')
            nn = ASTNestedName([ASTIdentifier(self.matched_text)], rooted=False)
            # Allow named variadic args:
            # https://gcc.gnu.org/onlinedocs/cpp/Variadic-Macros.html
            self.skip_ws()
            if self.skip_string_and_ws('...'):
                args.append(ASTMacroParameter(nn, False, True))
                self.skip_ws()
                if not self.skip_string(')'):
                    self.fail('Expected ")" after "..." in macro parameters.')
                break
            args.append(ASTMacroParameter(nn))
            if self.skip_string_and_ws(','):
                continue
            if self.skip_string_and_ws(')'):
                break
            self.fail("Expected identifier, ')', or ',' in macro parameter list.")
        return ASTMacro(ident, args)

    def _parse_struct(self) -> ASTStruct:
        name = self._parse_nested_name()
        return ASTStruct(name)

    def _parse_union(self) -> ASTUnion:
        name = self._parse_nested_name()
        return ASTUnion(name)

    def _parse_enum(self) -> ASTEnum:
        name = self._parse_nested_name()
        return ASTEnum(name)

    def _parse_enumerator(self) -> ASTEnumerator:
        name = self._parse_nested_name()
        attrs = self._parse_attribute_list()
        self.skip_ws()
        init = None
        if self.skip_string('='):
            self.skip_ws()
            parser = self._parse_constant_expression
            init_val = self._parse_expression_fallback([], parser)
            init = ASTInitializer(init_val)
        return ASTEnumerator(name, init, attrs)

    def parse_declaration(self, objectType: str, directiveType: str) -> ASTDeclaration:
        object_type = objectType
        directive_type = directiveType
        if object_type not in {
            'function',
            'member',
            'macro',
            'struct',
            'union',
            'enum',
            'enumerator',
            'type',
        }:
            raise Exception('Internal error, unknown objectType "%s".' % object_type)
        if directive_type not in {
            'function',
            'member',
            'var',
            'macro',
            'struct',
            'union',
            'enum',
            'enumerator',
            'type',
        }:
            raise Exception(
                'Internal error, unknown directiveType "%s".' % directive_type
            )

        declaration: DeclarationType | None = None
        if object_type == 'member':
            declaration = self._parse_type_with_init(named=True, outer='member')
        elif object_type == 'function':
            declaration = self._parse_type(named=True, outer='function')
        elif object_type == 'macro':
            declaration = self._parse_macro()
        elif object_type == 'struct':
            declaration = self._parse_struct()
        elif object_type == 'union':
            declaration = self._parse_union()
        elif object_type == 'enum':
            declaration = self._parse_enum()
        elif object_type == 'enumerator':
            declaration = self._parse_enumerator()
        elif object_type == 'type':
            declaration = self._parse_type(named=True, outer='type')
        else:
            raise AssertionError
        if object_type != 'macro':
            self.skip_ws()
            semicolon = self.skip_string(';')
        else:
            semicolon = False
        return ASTDeclaration(object_type, directive_type, declaration, semicolon)

    def parse_namespace_object(self) -> ASTNestedName:
        return self._parse_nested_name()

    def parse_xref_object(self) -> ASTNestedName:
        name = self._parse_nested_name()
        # if there are '()' left, just skip them
        self.skip_ws()
        self.skip_string('()')
        self.assert_end()
        return name

    def parse_expression(self) -> ASTExpression | ASTType:
        pos = self.pos
        res: ASTExpression | ASTType | None = None
        try:
            res = self._parse_expression()
            self.skip_ws()
            self.assert_end()
        except DefinitionError as ex_expr:
            self.pos = pos
            try:
                res = self._parse_type(False)
                self.skip_ws()
                self.assert_end()
            except DefinitionError as ex_type:
                header = 'Error when parsing (type) expression.'
                errs = [
                    (ex_expr, 'If expression'),
                    (ex_type, 'If type'),
                ]
                raise self._make_multi_error(errs, header) from ex_type
        return res



"""Docutils transforms used by Sphinx."""

from __future__ import annotations

import os
import re
from hashlib import sha1
from math import ceil
from pathlib import Path
from typing import TYPE_CHECKING

from docutils import nodes

from sphinx.locale import __
from sphinx.transforms import SphinxTransform
from sphinx.util import logging, requests
from sphinx.util._pathlib import _StrPath
from sphinx.util.http_date import epoch_to_rfc1123, rfc1123_to_epoch
from sphinx.util.images import get_image_extension, guess_mimetype, parse_data_uri
from sphinx.util.osutil import ensuredir

if TYPE_CHECKING:
    from typing import Any

    from sphinx.application import Sphinx
    from sphinx.util.typing import ExtensionMetadata

logger = logging.getLogger(__name__)

MAX_FILENAME_LEN = 32
CRITICAL_PATH_CHAR_RE = re.compile('[:;<>|*" ]')


class BaseImageConverter(SphinxTransform):
    def apply(self, **kwargs: Any) -> None:
        for node in self.document.findall(nodes.image):
            if self.match(node):
                self.handle(node)

    def match(self, node: nodes.image) -> bool:
        return True

    def handle(self, node: nodes.image) -> None:
        pass

    @property
    def imagedir(self) -> _StrPath:
        return self.env.doctreedir / 'images'


class ImageDownloader(BaseImageConverter):
    default_priority = 100

    def match(self, node: nodes.image) -> bool:
        if not self.env._builder_cls.supported_image_types:
            return False
        if self.env._builder_cls.supported_remote_images:
            return False
        return '://' in node['uri']

    def handle(self, node: nodes.image) -> None:
        try:
            basename = os.path.basename(node['uri'])
            if '?' in basename:
                basename = basename.split('?')[0]
            if not basename or len(basename) > MAX_FILENAME_LEN:
                filename, ext = os.path.splitext(node['uri'])
                basename = (
                    sha1(filename.encode(), usedforsecurity=False).hexdigest() + ext
                )
            basename = CRITICAL_PATH_CHAR_RE.sub('_', basename)

            uri_hash = sha1(node['uri'].encode(), usedforsecurity=False).hexdigest()
            path = Path(self.imagedir, uri_hash, basename)
            path.parent.mkdir(parents=True, exist_ok=True)
            self._download_image(node, path)

        except Exception as exc:
            msg = __('Could not fetch remote image: %s [%s]')
            logger.warning(msg, node['uri'], exc)

    def _download_image(self, node: nodes.image, path: Path) -> None:
        headers = {}
        if path.exists():
            timestamp: float = ceil(path.stat().st_mtime)
            headers['If-Modified-Since'] = epoch_to_rfc1123(timestamp)

        config = self.config
        r = requests.get(
            node['uri'],
            headers=headers,
            _user_agent=config.user_agent,
            _tls_info=(config.tls_verify, config.tls_cacerts),
        )
        if r.status_code >= 400:
            msg = __('Could not fetch remote image: %s [%d]')
            logger.warning(msg, node['uri'], r.status_code)
        else:
            self.env.original_image_uri[_StrPath(path)] = node['uri']

            if r.status_code == 200:
                path.write_bytes(r.content)
            if last_modified := r.headers.get('Last-Modified'):
                timestamp = rfc1123_to_epoch(last_modified)
                os.utime(path, (timestamp, timestamp))

            self._process_image(node, path)

    def _process_image(self, node: nodes.image, path: Path) -> None:
        str_path = _StrPath(path)
        self.env.original_image_uri[str_path] = node['uri']

        mimetype = guess_mimetype(path, default='*')
        if mimetype != '*' and not path.suffix:
            # append a suffix if URI does not contain suffix
            ext = get_image_extension(mimetype) or ''
            with_ext = path.with_name(path.name + ext)
            path.replace(with_ext)
            self.env.original_image_uri.pop(str_path)
            self.env.original_image_uri[_StrPath(with_ext)] = node['uri']
            path = with_ext
        path_str = str(path)
        node['candidates'].pop('?')
        node['candidates'][mimetype] = path_str
        node['uri'] = path_str
        self.env.images.add_file(self.env.current_document.docname, path_str)


class DataURIExtractor(BaseImageConverter):
    default_priority = 150

    def match(self, node: nodes.image) -> bool:
        if self.env._builder_cls.supported_data_uri_images is True:
            return False  # do not transform the image; data URIs are valid in the build output
        return node['uri'].startswith('data:')

    def handle(self, node: nodes.image) -> None:
        image = parse_data_uri(node['uri'])
        assert image is not None
        ext = get_image_extension(image.mimetype)
        if ext is None:
            logger.warning(
                __('Unknown image format: %s...'), node['uri'][:32], location=node
            )
            return

        ensuredir(self.imagedir / 'embeded')
        digest = sha1(image.data, usedforsecurity=False).hexdigest()
        path = self.imagedir / 'embeded' / (digest + ext)
        self.env.original_image_uri[path] = node['uri']

        with open(path, 'wb') as f:
            f.write(image.data)

        path_str = str(path)
        node['candidates'].pop('?')
        node['candidates'][image.mimetype] = path_str
        node['uri'] = path_str
        self.env.images.add_file(self.env.current_document.docname, path_str)


def get_filename_for(filename: str, mimetype: str) -> str:
    basename = os.path.basename(filename)
    basename = CRITICAL_PATH_CHAR_RE.sub('_', basename)
    return os.path.splitext(basename)[0] + (get_image_extension(mimetype) or '')


class ImageConverter(BaseImageConverter):
    """A base class for image converters.

    An image converter is kind of Docutils transform module.  It is used to
    convert image files which are not supported by a builder to the
    appropriate format for that builder.

    For example, :py:class:`LaTeX builder <.LaTeXBuilder>` supports PDF,
    PNG and JPEG as image formats.  However it does not support SVG images.
    For such case, using image converters allows to embed these
    unsupported images into the document.  One of the image converters;
    :ref:`sphinx.ext.imgconverter <sphinx.ext.imgconverter>` can convert
    a SVG image to PNG format using Imagemagick internally.

    There are three steps to make your custom image converter:

    1. Make a subclass of ``ImageConverter`` class
    2. Override ``conversion_rules``, ``is_available()`` and ``convert()``
    3. Register your image converter to Sphinx using
       :py:meth:`.Sphinx.add_post_transform`
    """

    default_priority = 200

    #: The converter is available or not.  Will be filled at the first call of
    #: the build.  The result is shared in the same process.
    #:
    #: .. todo:: This should be refactored not to store the state without class
    #:           variable.
    available: bool | None = None

    #: A conversion rules the image converter supports.
    #: It is represented as a list of pair of source image format (mimetype) and
    #: destination one::
    #:
    #:     conversion_rules = [
    #:         ('image/svg+xml', 'image/png'),
    #:         ('image/gif', 'image/png'),
    #:         ('application/pdf', 'image/png'),
    #:     ]
    conversion_rules: list[tuple[str, str]] = []

    def match(self, node: nodes.image) -> bool:
        if not self.env._builder_cls.supported_image_types:
            return False
        if '?' in node['candidates']:
            return False
        node_mime_types = set(self.guess_mimetypes(node))
        supported_image_types = set(self.env._builder_cls.supported_image_types)
        if node_mime_types & supported_image_types:
            # builder supports the image; no need to convert
            return False
        if self.available is None:
            # store the value to the class variable to share it during the build
            self.__class__.available = self.is_available()

        if not self.available:
            return False
        else:
            try:
                self.get_conversion_rule(node)
            except ValueError:
                return False
            else:
                return True

    def get_conversion_rule(self, node: nodes.image) -> tuple[str, str]:
        for candidate in self.guess_mimetypes(node):
            for supported in self.env._builder_cls.supported_image_types:
                rule = (candidate, supported)
                if rule in self.conversion_rules:
                    return rule

        msg = 'No conversion rule found'
        raise ValueError(msg)

    def is_available(self) -> bool:
        """Return the image converter is available or not."""
        raise NotImplementedError

    def guess_mimetypes(self, node: nodes.image) -> list[str]:
        # The special key ? is set for nonlocal URIs.
        if '?' in node['candidates']:
            return []
        elif '*' in node['candidates']:
            path = self.env.srcdir / node['uri']
            guessed = guess_mimetype(path)
            return [guessed] if guessed is not None else []
        else:
            return node['candidates'].keys()

    def handle(self, node: nodes.image) -> None:
        _from, _to = self.get_conversion_rule(node)

        if _from in node['candidates']:
            srcpath = node['candidates'][_from]
        else:
            srcpath = node['candidates']['*']

        filename = self.env.images[srcpath][1]
        filename = get_filename_for(filename, _to)
        ensuredir(self.imagedir)
        destpath = self.imagedir / filename

        abs_srcpath = self.env.srcdir / srcpath
        if self.convert(abs_srcpath, destpath):
            if '*' in node['candidates']:
                node['candidates']['*'] = str(destpath)
            else:
                node['candidates'][_to] = str(destpath)
            node['uri'] = str(destpath)

            self.env.original_image_uri[destpath] = srcpath
            self.env.images.add_file(self.env.current_document.docname, destpath)

    def convert(
        self, _from: str | os.PathLike[str], _to: str | os.PathLike[str]
    ) -> bool:
        """Convert an image file to the expected format.

        *_from* is a path of the source image file, and *_to* is a path
        of the destination file.
        """
        raise NotImplementedError


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_post_transform(ImageDownloader)
    app.add_post_transform(DataURIExtractor)

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Sphinx theme specific highlighting styles."""

from __future__ import annotations

from pygments.style import Style
from pygments.styles.friendly import FriendlyStyle
from pygments.token import (
    Comment,
    Error,
    Generic,
    Keyword,
    Name,
    Number,
    Operator,
    String,
    Whitespace,
)


class NoneStyle(Style):
    """Style without any styling."""


class SphinxStyle(Style):
    """Like friendly, but a bit darker to enhance contrast on the green
    background.
    """

    background_color = '#eeffcc'
    default_style = ''

    styles = {
        **FriendlyStyle.styles,
        Generic.Output: '#333',
        Comment: 'italic #408090',
        Number: '#208050',
    }


class PyramidStyle(Style):
    """Pylons/pyramid pygments style based on friendly style, by Blaise Laflamme."""

    # work in progress...

    background_color = '#f8f8f8'
    default_style = ''

    styles = {
        Whitespace:         '#bbbbbb',
        Comment:            'italic #60a0b0',
        Comment.Preproc:    'noitalic #007020',
        Comment.Special:    'noitalic bg:#fff0f0',

        Keyword:            'bold #007020',
        Keyword.Pseudo:     'nobold',
        Keyword.Type:       'nobold #902000',

        Operator:           '#666666',
        Operator.Word:      'bold #007020',

        Name.Builtin:       '#007020',
        Name.Function:      '#06287e',
        Name.Class:         'bold #0e84b5',
        Name.Namespace:     'bold #0e84b5',
        Name.Exception:     '#007020',
        Name.Variable:      '#bb60d5',
        Name.Constant:      '#60add5',
        Name.Label:         'bold #002070',
        Name.Entity:        'bold #d55537',
        Name.Attribute:     '#0e84b5',
        Name.Tag:           'bold #062873',
        Name.Decorator:     'bold #555555',

        String:             '#4070a0',
        String.Doc:         'italic',
        String.Interpol:    'italic #70a0d0',
        String.Escape:      'bold #4070a0',
        String.Regex:       '#235388',
        String.Symbol:      '#517918',
        String.Other:       '#c65d09',
        Number:             '#40a070',

        Generic.Heading:    'bold #000080',
        Generic.Subheading: 'bold #800080',
        Generic.Deleted:    '#A00000',
        Generic.Inserted:   '#00A000',
        Generic.Error:      '#FF0000',
        Generic.Emph:       'italic',
        Generic.Strong:     'bold',
        Generic.Prompt:     'bold #c65d09',
        Generic.Output:     '#888',
        Generic.Traceback:  '#04D',

        Error:              '#a40000 bg:#fbe3e4',
    }  # fmt: skip



"""Handlers for additional ReST directives."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, cast

from docutils import nodes
from docutils.parsers.rst import directives, roles

from sphinx import addnodes
from sphinx.util import docutils
from sphinx.util.docfields import DocFieldTransformer
from sphinx.util.docutils import SphinxDirective

if TYPE_CHECKING:
    from typing import ClassVar

    from docutils.nodes import Node

    from sphinx.addnodes import desc_signature
    from sphinx.application import Sphinx
    from sphinx.util.docfields import Field, TypedField
    from sphinx.util.typing import ExtensionMetadata, OptionSpec


# RE to strip backslash escapes
nl_escape_re = re.compile(r'\\\n')
strip_backslash_re = re.compile(r'\\(.)')


def optional_int(argument: str) -> int | None:
    """Check for an integer argument or None value; raise ``ValueError`` if not."""
    if argument is None:
        return None
    else:
        value = int(argument)
        if value < 0:
            msg = 'negative value; must be positive or zero'
            raise ValueError(msg)
        return value


class ObjectDescription[ObjDescT](SphinxDirective):
    """Directive to describe a class, function or similar object.

    Not used directly, but subclassed (in domain-specific directives)
    to add custom behaviour.
    """

    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec: ClassVar[OptionSpec] = {
        'no-index': directives.flag,
        'no-index-entry': directives.flag,
        'no-contents-entry': directives.flag,
        'no-typesetting': directives.flag,
        'noindex': directives.flag,
        'noindexentry': directives.flag,
        'nocontentsentry': directives.flag,
    }

    # types of doc fields that this directive handles, see sphinx.util.docfields
    doc_field_types: list[Field] = []
    domain: str | None = None
    objtype: str  # set when `run` method is called
    indexnode: addnodes.index

    # Warning: this might be removed in future version. Don't touch this from extensions.
    _doc_field_type_map: dict[str, tuple[Field, bool]] = {}

    def get_field_type_map(self) -> dict[str, tuple[Field, bool]]:
        if self._doc_field_type_map == {}:
            self._doc_field_type_map = {}
            for field in self.doc_field_types:
                for name in field.names:
                    self._doc_field_type_map[name] = (field, False)

                if field.is_typed:
                    typed_field = cast('TypedField', field)
                    for name in typed_field.typenames:
                        self._doc_field_type_map[name] = (field, True)

        return self._doc_field_type_map

    def get_signatures(self) -> list[str]:
        """Retrieve the signatures to document from the directive arguments.

        By default, signatures are given as arguments, one per line.
        """
        lines = nl_escape_re.sub('', self.arguments[0]).split('\n')
        if self.config.strip_signature_backslash:
            # remove backslashes to support (dummy) escapes; helps Vim highlighting
            return [strip_backslash_re.sub(r'\1', line.strip()) for line in lines]
        else:
            return [line.strip() for line in lines]

    def handle_signature(self, sig: str, signode: desc_signature) -> ObjDescT:
        """Parse the signature *sig*.

        The individual nodes are then appended to *signode*.
        If ValueError is raised, parsing is aborted and the whole
        *sig* is put into a single desc_name node.

        The return value should be a value that identifies the object.  It is
        passed to :meth:`add_target_and_index()` unchanged, and otherwise only
        used to skip duplicates.
        """
        raise ValueError

    def add_target_and_index(
        self, name: ObjDescT, sig: str, signode: desc_signature
    ) -> None:
        """Add cross-reference IDs and entries to self.indexnode, if applicable.

        *name* is whatever :meth:`handle_signature()` returned.
        """
        pass  # do nothing by default

    def before_content(self) -> None:
        """Called before parsing content.

        Used to set information about the current directive context
        on the build environment.
        """
        pass

    def transform_content(self, content_node: addnodes.desc_content) -> None:
        """Can be used to manipulate the content.

        Called after creating the content through nested parsing,
        but before the ``object-description-transform`` event is emitted,
        and before the info-fields are transformed.
        """
        pass

    def after_content(self) -> None:
        """Called after parsing content.

        Used to reset information about the current directive context
        on the build environment.
        """
        pass

    def _object_hierarchy_parts(self, sig_node: desc_signature) -> tuple[str, ...]:
        """Returns a tuple of strings, one entry for each part of the object's
        hierarchy (e.g. ``('module', 'submodule', 'Class', 'method')``). The
        returned tuple is used to properly nest children within parents in the
        table of contents, and can also be used within the
        :py:meth:`_toc_entry_name` method.

        This method must not be used outwith table of contents generation.
        """
        return ()

    def _toc_entry_name(self, sig_node: desc_signature) -> str:
        """Returns the text of the table of contents entry for the object.

        This function is called once, in :py:meth:`run`, to set the name for the
        table of contents entry (a special attribute ``_toc_name`` is set on the
        object node, later used in
        ``environment.collectors.toctree.TocTreeCollector.process_doc().build_toc()``
        when the table of contents entries are collected).

        To support table of contents entries for their objects, domains must
        override this method, also respecting the configuration setting
        ``toc_object_entries_show_parents``. Domains must also override
        :py:meth:`_object_hierarchy_parts`, with one (string) entry for each part of the
        object's hierarchy. The result of this method is set on the signature
        node, and can be accessed as ``sig_node['_toc_parts']`` for use within
        this method. The resulting tuple is also used to properly nest children
        within parents in the table of contents.

        An example implementations of this method is within the python domain
        (:meth:`!PyObject._toc_entry_name`). The python domain sets the
        ``_toc_parts`` attribute within the :py:meth:`handle_signature()`
        method.
        """
        return ''

    def run(self) -> list[Node]:
        """Main directive entry function, called by docutils upon encountering the
        directive.

        This directive is meant to be quite easily subclassable, so it delegates
        to several additional methods.  What it does:

        * find out if called as a domain-specific directive, set self.domain
        * create a `desc` node to fit all description inside
        * parse standard options, currently `no-index`
        * create an index node if needed as self.indexnode
        * parse all given signatures (as returned by self.get_signatures())
          using self.handle_signature(), which should either return a name
          or raise ValueError
        * add index entries using self.add_target_and_index()
        * parse the content and handle doc fields in it
        """
        if ':' in self.name:
            self.domain, _, self.objtype = self.name.partition(':')
        else:
            self.domain, self.objtype = '', self.name
        self.indexnode = addnodes.index(entries=[])

        node = addnodes.desc()
        node.document = self.state.document
        source, line = self.get_source_info()
        # If any options were specified to the directive,
        # self.state.document.current_line will at this point be set to
        # None.  To ensure nodes created as part of the signature have a line
        # number set, set the document's line number correctly.
        #
        # Note that we need to subtract one from the line number since
        # note_source uses 0-based line numbers.
        if line is not None:
            line -= 1
        self.state.document.note_source(source, line)  # type: ignore[arg-type]
        node['domain'] = self.domain
        # 'desctype' is a backwards compatible attribute
        node['objtype'] = node['desctype'] = self.objtype

        # Copy old option names to new ones
        # xref RemovedInSphinx90Warning
        # deprecate noindex, noindexentry, and nocontentsentry in Sphinx 9.0
        if 'no-index' not in self.options and 'noindex' in self.options:
            self.options['no-index'] = self.options['noindex']
        if 'no-index-entry' not in self.options and 'noindexentry' in self.options:
            self.options['no-index-entry'] = self.options['noindexentry']
        if (
            'no-contents-entry' not in self.options
            and 'nocontentsentry' in self.options
        ):
            self.options['no-contents-entry'] = self.options['nocontentsentry']

        node['no-index'] = node['noindex'] = no_index = 'no-index' in self.options
        node['no-index-entry'] = node['noindexentry'] = 'no-index-entry' in self.options
        node['no-contents-entry'] = node['nocontentsentry'] = (
            'no-contents-entry' in self.options
        )
        node['no-typesetting'] = 'no-typesetting' in self.options
        if self.domain:
            node['classes'].append(self.domain)
        node['classes'].append(node['objtype'])

        self.names: list[ObjDescT] = []
        signatures = self.get_signatures()
        for sig in signatures:
            # add a signature node for each signature in the current unit
            # and add a reference target for it
            signode = addnodes.desc_signature(sig, '')
            self.set_source_info(signode)
            node.append(signode)
            try:
                # name can also be a tuple, e.g. (classname, objname);
                # this is strictly domain-specific (i.e. no assumptions may
                # be made in this base class)
                name = self.handle_signature(sig, signode)
            except ValueError:
                # signature parsing failed
                signode.clear()
                signode += addnodes.desc_name(sig, sig)
                continue  # we don't want an index entry here
            finally:
                # Private attributes for ToC generation. Will be modified or removed
                # without notice.
                if self.config.toc_object_entries:
                    signode['_toc_parts'] = self._object_hierarchy_parts(signode)
                    signode['_toc_name'] = self._toc_entry_name(signode)
                else:
                    signode['_toc_parts'] = ()
                    signode['_toc_name'] = ''
            if name not in self.names:
                self.names.append(name)
                if not no_index:
                    # only add target and index entry if this is the first
                    # description of the object with this name in this desc block
                    self.add_target_and_index(name, sig, signode)

        if self.names:
            # needed for association of version{added,changed} directives
            object_name: ObjDescT = self.names[0]
            if isinstance(object_name, tuple):
                self.env.current_document.obj_desc_name = str(object_name[0])
            else:
                self.env.current_document.obj_desc_name = str(object_name)
        self.before_content()
        content_children = self.parse_content_to_nodes(allow_section_headings=True)
        content_node = addnodes.desc_content('', *content_children)
        node.append(content_node)
        self.transform_content(content_node)
        self.env.events.emit(
            'object-description-transform', self.domain, self.objtype, content_node
        )
        DocFieldTransformer(self).transform_all(content_node)
        self.env.current_document.obj_desc_name = ''
        self.after_content()

        if node['no-typesetting']:
            # Attempt to return the index node, and a new target node
            # containing all the ids of this node and its children.
            # If ``:no-index:`` is set, or there are no ids on the node
            # or any of its children, then just return the index node,
            # as Docutils expects a target node to have at least one id.
            if node_ids := [  # type: ignore[var-annotated]
                node_id
                for el in node.findall(nodes.Element)
                for node_id in el.get('ids', ())
            ]:
                target_node = nodes.target(ids=node_ids)
                self.set_source_info(target_node)
                return [self.indexnode, target_node]
            return [self.indexnode]
        return [self.indexnode, node]


class DefaultRole(SphinxDirective):
    """Set the default interpreted text role.  Overridden from docutils."""

    optional_arguments = 1
    final_argument_whitespace = False

    def run(self) -> list[Node]:
        if not self.arguments:
            docutils.unregister_role('')
            return []
        role_name = self.arguments[0]
        # TODO: TYPING: Upstream docutils should widen roles.role() to accept
        #       the RST language module shape (has a `roles` mapping);
        #       current stub requires _LanguageModule from docutils.languages.
        role, messages = roles.role(
            role_name,
            self.state_machine.language,  # type: ignore[arg-type]
            self.lineno,
            self.state.reporter,
        )
        if role:
            docutils.register_role('', role)  # type: ignore[arg-type]
            self.env.current_document.default_role = role_name
        else:
            literal_block = nodes.literal_block(self.block_text, self.block_text)
            reporter = self.state.reporter
            error = reporter.error(
                'Unknown interpreted text role "%s".' % role_name,
                literal_block,
                line=self.lineno,
            )
            messages += [error]

        return cast('list[nodes.Node]', messages)


class DefaultDomain(SphinxDirective):
    """Directive to (re-)set the default domain for this source file."""

    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec: ClassVar[OptionSpec] = {}

    def run(self) -> list[Node]:
        domain_name = self.arguments[0].lower()
        default_domain = self.env.domains.get(domain_name)
        self.env.current_document.default_domain = default_domain
        return []


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_config_value(
        'strip_signature_backslash', False, 'env', types=frozenset({bool})
    )
    directives.register_directive('default-role', DefaultRole)
    directives.register_directive('default-domain', DefaultDomain)
    directives.register_directive('describe', ObjectDescription)
    # new, more consistent, name
    directives.register_directive('object', ObjectDescription)

    app.add_event('object-description-transform')

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Convert times to and from HTTP-date serialisations.

Reference: https://www.rfc-editor.org/rfc/rfc7231#section-7.1.1.1
"""

from __future__ import annotations

import time
from email.utils import parsedate_tz

_WEEKDAY_NAME = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')
_MONTH_NAME = ('',  # Placeholder for indexing purposes
               'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
               'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')  # fmt: skip
_GMT_OFFSET = float(time.localtime().tm_gmtoff)


def epoch_to_rfc1123(epoch: float) -> str:
    """Return HTTP-date string from epoch offset."""
    yr, mn, dd, hh, mm, ss, wd, _yd, _tz = time.gmtime(epoch)
    weekday_name = _WEEKDAY_NAME[wd]
    month = _MONTH_NAME[mn]
    return f'{weekday_name}, {dd:02} {month} {yr:04} {hh:02}:{mm:02}:{ss:02} GMT'


def rfc1123_to_epoch(rfc1123: str) -> float:
    """Return epoch offset from HTTP-date string."""
    t = parsedate_tz(rfc1123)
    if t is None:
        raise ValueError
    if not rfc1123.endswith(' GMT'):
        msg = (
            'HTTP-date string does not meet RFC 7231 requirements '
            f"(must end with 'GMT'): {rfc1123!r}"
        )
        raise ValueError(msg)
    gmt_offset = t[9]
    if gmt_offset != 0:
        msg = (
            'HTTP-date string does not meet RFC 7231 requirements '
            f'(must be GMT time): {rfc1123!r}'
        )
        raise ValueError(msg)
    epoch_secs = time.mktime(time.struct_time(t[:9])) + _GMT_OFFSET
    return epoch_secs



"""mock for autodoc"""

from __future__ import annotations

import contextlib
import os
import sys
from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec
from types import MethodType, ModuleType
from typing import TYPE_CHECKING

from sphinx.ext.autodoc._shared import LOGGER
from sphinx.util.inspect import isboundmethod, safe_getattr

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence, Set
    from typing import Any

    from typing_extensions import TypeIs


class _MockObject:
    """Used by autodoc_mock_imports."""

    __display_name__ = '_MockObject'
    __name__ = ''
    __sphinx_mock__ = True
    __sphinx_decorator_args__: tuple[Any, ...] = ()
    # Attributes listed here should not be mocked and rather raise an Attribute error:
    __sphinx_empty_attrs__: Set[str] = frozenset(('__typing_subst__',))

    def __new__(cls, *args: Any, **kwargs: Any) -> Any:  # NoQA: ARG004
        if len(args) == 3 and isinstance(args[1], tuple):
            superclass = args[1][-1].__class__
            if superclass is cls:
                # subclassing MockObject
                return _make_subclass(
                    args[0],
                    superclass.__display_name__,
                    superclass=superclass,
                    attributes=args[2],
                )

        return super().__new__(cls)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.__qualname__ = self.__name__

    def __len__(self) -> int:
        return 0

    def __contains__(self, key: str) -> bool:
        return False

    def __iter__(self) -> Iterator[Any]:
        return iter(())

    def __mro_entries__(self, bases: tuple[Any, ...]) -> tuple[type, ...]:
        return (self.__class__,)

    def __getitem__(self, key: Any) -> _MockObject:
        return _make_subclass(str(key), self.__display_name__, self.__class__)()

    def __getattr__(self, key: str) -> _MockObject:
        if key in self.__sphinx_empty_attrs__:
            raise AttributeError
        return _make_subclass(key, self.__display_name__, self.__class__)()

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        call = self.__class__()
        call.__sphinx_decorator_args__ = args
        return call

    def __repr__(self) -> str:
        return self.__display_name__


def _make_subclass(
    name: str,
    module: str,
    superclass: Any = _MockObject,
    attributes: Any = None,
    decorator_args: tuple[Any, ...] = (),
) -> Any:
    attrs = {
        '__module__': module,
        '__display_name__': module + '.' + name,
        '__name__': name,
        '__sphinx_decorator_args__': decorator_args,
    }
    attrs.update(attributes or {})

    return type(name, (superclass,), attrs)


class _MockModule(ModuleType):
    """Used by autodoc_mock_imports."""

    __file__ = os.devnull
    __sphinx_mock__ = True

    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.__all__: list[str] = []
        self.__path__: list[str] = []

    def __getattr__(self, name: str) -> _MockObject:
        return _make_subclass(name, self.__name__)()

    def __repr__(self) -> str:
        return self.__name__


class MockLoader(Loader):
    """A loader for mocking."""

    def __init__(self, finder: MockFinder) -> None:
        super().__init__()
        self.finder = finder

    def create_module(self, spec: ModuleSpec) -> ModuleType:
        LOGGER.debug('[autodoc] adding a mock module as %s!', spec.name)
        self.finder.mocked_modules.append(spec.name)
        return _MockModule(spec.name)

    def exec_module(self, module: ModuleType) -> None:
        pass  # nothing to do


class MockFinder(MetaPathFinder):
    """A finder for mocking."""

    def __init__(self, modnames: Sequence[str]) -> None:
        super().__init__()
        self.modnames = modnames
        self.loader = MockLoader(self)
        self.mocked_modules: list[str] = []

    def find_spec(
        self,
        fullname: str,
        path: Sequence[bytes | str] | None,
        target: ModuleType | None = None,
    ) -> ModuleSpec | None:
        for modname in self.modnames:
            # check if fullname is (or is a descendant of) one of our targets
            if modname == fullname or fullname.startswith(modname + '.'):
                return ModuleSpec(fullname, self.loader)

        return None

    def invalidate_caches(self) -> None:
        """Invalidate mocked modules on sys.modules."""
        for modname in self.mocked_modules:
            sys.modules.pop(modname, None)


@contextlib.contextmanager
def mock(modnames: Sequence[str]) -> Iterator[None]:
    """Insert mock modules during context::

    with mock(['target.module.name']):
        # mock modules are enabled here
        ...
    """
    finder = MockFinder(modnames)
    try:
        sys.meta_path.insert(0, finder)
        yield
    finally:
        sys.meta_path.remove(finder)
        finder.invalidate_caches()


def ismockmodule(subject: Any) -> TypeIs[_MockModule]:
    """Check if the object is a mocked module."""
    return isinstance(subject, _MockModule)


def ismock(subject: Any) -> bool:
    """Check if the object is mocked."""
    # check the object has '__sphinx_mock__' attribute
    try:
        if safe_getattr(subject, '__sphinx_mock__', None) is None:
            return False
    except AttributeError:
        return False

    # check the object is mocked module
    if isinstance(subject, _MockModule):
        return True

    # check the object is bound method
    if isinstance(subject, MethodType) and isboundmethod(subject):
        tmp_subject = subject.__func__
    else:
        tmp_subject = subject

    try:
        # check the object is mocked object
        __mro__ = safe_getattr(type(tmp_subject), '__mro__', [])
        if len(__mro__) > 2 and __mro__[-2] is _MockObject:
            # A mocked object has a MRO that ends with (..., _MockObject, object).
            return True
    except AttributeError:
        pass

    return False


def undecorate(subject: _MockObject) -> Any:
    """Unwrap mock if *subject* is decorated by mocked object.

    If not decorated, returns given *subject* itself.
    """
    if ismock(subject) and subject.__sphinx_decorator_args__:
        return subject.__sphinx_decorator_args__[0]
    else:
        return subject



"""The image collector for sphinx.environment."""

from __future__ import annotations

import os
import os.path
from glob import glob
from pathlib import Path
from typing import TYPE_CHECKING

from docutils import nodes

from sphinx import addnodes
from sphinx.environment.collectors import EnvironmentCollector
from sphinx.locale import __
from sphinx.util import logging
from sphinx.util.i18n import get_image_filename_for_language, search_image_for_language
from sphinx.util.images import guess_mimetype
from sphinx.util.osutil import _relative_path

if TYPE_CHECKING:
    from collections.abc import Set

    from docutils.nodes import Node

    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment
    from sphinx.util.typing import ExtensionMetadata

logger = logging.getLogger(__name__)


class ImageCollector(EnvironmentCollector):
    """Image files collector for sphinx.environment."""

    def clear_doc(self, app: Sphinx, env: BuildEnvironment, docname: str) -> None:
        env.images.purge_doc(docname)

    def merge_other(
        self,
        app: Sphinx,
        env: BuildEnvironment,
        docnames: Set[str],
        other: BuildEnvironment,
    ) -> None:
        env.images.merge_other(docnames, other.images)

    def process_doc(self, app: Sphinx, doctree: nodes.document) -> None:
        """Process and rewrite image URIs."""
        docname = app.env.current_document.docname

        for node in doctree.findall(nodes.image):
            # Map the mimetype to the corresponding image.  The writer may
            # choose the best image from these candidates.  The special key * is
            # set if there is only single candidate to be used by a writer.
            # The special key ? is set for nonlocal URIs.
            candidates: dict[str, str] = {}
            node['candidates'] = candidates
            imguri = node['uri']
            if imguri.startswith('data:'):
                candidates['?'] = imguri
                continue
            if imguri.find('://') != -1:
                candidates['?'] = imguri
                continue

            if imguri.endswith(os.extsep + '*'):
                # Update `node['uri']` to a relative path from srcdir
                # from a relative path from current document.
                rel_imgpath, full_imgpath = app.env.relfn2path(imguri, docname)
                node['uri'] = rel_imgpath

                # Search language-specific figures at first
                i18n_imguri = get_image_filename_for_language(imguri, app.env)
                _, full_i18n_imgpath = app.env.relfn2path(i18n_imguri, docname)
                self.collect_candidates(app.srcdir, full_i18n_imgpath, candidates, node)

                self.collect_candidates(app.srcdir, full_imgpath, candidates, node)
            else:
                # substitute imguri by figure_language_filename
                # (ex. foo.png -> foo.en.png)
                imguri = search_image_for_language(imguri, app.env)

                # Update `node['uri']` to a relative path from srcdir
                # from a relative path from current document.
                original_uri = node['uri']
                node['uri'], _ = app.env.relfn2path(imguri, docname)
                candidates['*'] = node['uri']
                if node['uri'] != original_uri:
                    node['original_uri'] = original_uri

            # map image paths to unique image names (so that they can be put
            # into a single directory)
            for imgpath in candidates.values():
                app.env.note_dependency(imgpath)
                if not os.access(app.srcdir / imgpath, os.R_OK):
                    logger.warning(
                        __('image file not readable: %s'),
                        imgpath,
                        location=node,
                        type='image',
                        subtype='not_readable',
                    )
                    continue
                app.env.images.add_file(docname, imgpath)

    def collect_candidates(
        self,
        srcdir: Path,
        imgpath: str,
        candidates: dict[str, str],
        node: Node,
    ) -> None:
        globbed: dict[str, list[str]] = {}
        for filename in glob(imgpath):  # NoQA: PTH207
            new_imgpath = _relative_path(Path(filename), srcdir)
            try:
                mimetype = guess_mimetype(filename)
                if mimetype is None:
                    _basename, suffix = os.path.splitext(filename)
                    mimetype = 'image/x-' + suffix[1:]
                if mimetype not in candidates:
                    globbed.setdefault(mimetype, []).append(new_imgpath.as_posix())
            except OSError as err:
                logger.warning(
                    __('image file %s not readable: %s'),
                    filename,
                    err,
                    location=node,
                    type='image',
                    subtype='not_readable',
                )
        for key, files in globbed.items():
            candidates[key] = min(files, key=len)  # select by similarity


class DownloadFileCollector(EnvironmentCollector):
    """Download files collector for sphinx.environment."""

    def clear_doc(self, app: Sphinx, env: BuildEnvironment, docname: str) -> None:
        env.dlfiles.purge_doc(docname)

    def merge_other(
        self,
        app: Sphinx,
        env: BuildEnvironment,
        docnames: Set[str],
        other: BuildEnvironment,
    ) -> None:
        env.dlfiles.merge_other(docnames, other.dlfiles)

    def process_doc(self, app: Sphinx, doctree: nodes.document) -> None:
        """Process downloadable file paths."""
        for node in doctree.findall(addnodes.download_reference):
            targetname = node['reftarget']
            if '://' in targetname:
                node['refuri'] = targetname
            else:
                rel_filename, filename = app.env.relfn2path(
                    targetname, app.env.current_document.docname
                )
                app.env.note_dependency(rel_filename)
                if not os.access(filename, os.R_OK):
                    logger.warning(
                        __('download file not readable: %s'),
                        filename,
                        location=node,
                        type='download',
                        subtype='not_readable',
                    )
                    continue
                node['filename'] = app.env.dlfiles.add_file(
                    app.env.current_document.docname, rel_filename
                ).as_posix()


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_env_collector(ImageCollector)
    app.add_env_collector(DownloadFileCollector)

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""reStructuredText helper functions."""

from __future__ import annotations

import re
from collections import defaultdict
from contextlib import contextmanager
from typing import TYPE_CHECKING
from unicodedata import east_asian_width

from docutils.parsers.rst import roles
from docutils.parsers.rst.languages import en as english
from docutils.parsers.rst.states import Body
from docutils.utils import Reporter
from jinja2 import pass_environment

from sphinx.locale import __
from sphinx.util import docutils, logging

if TYPE_CHECKING:
    from collections.abc import Iterator

    from docutils.statemachine import StringList
    from jinja2 import Environment

logger = logging.getLogger(__name__)

_FIELD_NAME_RE = re.compile(Body.patterns['field_marker'])
symbols_re = re.compile(r'([!-\-/:-@\[-`{-~])')  # symbols without dot(0x2e)
SECTIONING_CHARS = ['=', '-', '~']

# width of characters
WIDECHARS: dict[str, str] = defaultdict(lambda: 'WF')  # WF: Wide + Full-width
WIDECHARS['ja'] = 'WFA'  # In Japanese, Ambiguous characters also have double width


def escape(text: str) -> str:
    text = symbols_re.sub(r'\\\1', text)
    text = re.sub(r'^\.', r'\.', text)  # escape a dot at top
    return text


def textwidth(text: str, widechars: str = 'WF') -> int:
    """Get width of text."""

    def charwidth(char: str, widechars: str) -> int:
        if east_asian_width(char) in widechars:
            return 2
        else:
            return 1

    return sum(charwidth(c, widechars) for c in text)


@pass_environment
def heading(env: Environment, text: str, level: int = 1) -> str:
    """Create a heading for *level*."""
    assert level <= 3
    # ``env.language`` is injected by ``sphinx.util.template.ReSTRenderer``
    width = textwidth(text, WIDECHARS[env.language])  # type: ignore[attr-defined]
    sectioning_char = SECTIONING_CHARS[level - 1]
    return f'{text}\n{sectioning_char * width}'


@contextmanager
def default_role(docname: str, name: str) -> Iterator[None]:
    if name:
        dummy_reporter = Reporter('', 4, 4)
        role_fn, _ = roles.role(name, english, 0, dummy_reporter)  # type: ignore[arg-type]
        if role_fn:
            docutils.register_role('', role_fn)  # type: ignore[arg-type]
        else:
            logger.warning(__('default role %s not found'), name, location=docname)

    yield

    docutils.unregister_role('')


def _prepend_prologue(content: StringList, prologue: str) -> None:
    """Prepend a string to content body as a prologue."""
    if not prologue:
        return
    pos = 0
    for line in content:
        if _FIELD_NAME_RE.match(line):
            pos += 1
        else:
            break

    if pos > 0:
        # insert a blank line after docinfo
        content.insert(pos, '', '<generated>', 0)
        pos += 1

    # insert prologue (after docinfo if exists)
    lineno = 0
    for lineno, line in enumerate(prologue.splitlines()):
        content.insert(pos + lineno, line, '<rst_prologue>', lineno)

    content.insert(pos + lineno + 1, '', '<generated>', 0)


def _append_epilogue(content: StringList, epilogue: str) -> None:
    """Append a string to content body as an epilogue."""
    if not epilogue:
        return
    if len(content) > 0:
        source, lineno = content.items[-1]
    else:
        source = '<generated>'
        lineno = 0
    content.append('', source, lineno + 1)
    for lineno, line in enumerate(epilogue.splitlines()):
        content.append(line, '<rst_epilogue>', lineno)



from __future__ import annotations

import dataclasses

from sphinx.ext.autodoc._sentinels import RUNTIME_INSTANCE_ATTRIBUTE, UNINITIALIZED_ATTR

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path
    from typing import Any, Literal

    from sphinx.ext.autodoc._sentinels import (
        RUNTIME_INSTANCE_ATTRIBUTE_T,
        SLOTS_ATTR_T,
        UNINITIALIZED_ATTR_T,
    )

    type _AutodocObjType = Literal[
        'module',
        'class',
        'exception',
        'function',
        'decorator',
        'method',
        'property',
        'attribute',
        'data',
        'type',
    ]
    type _AutodocFuncProperty = Literal[
        'abstractmethod',
        'async',
        'classmethod',
        'final',
        'singledispatch',
        'staticmethod',
    ]


@dataclasses.dataclass(frozen=False, kw_only=True, slots=True)
class _ItemProperties:
    #: The kind of object being documented
    obj_type: _AutodocObjType
    #: The dotted module name
    module_name: str
    #: The fully-qualified name within the module
    parts: tuple[str, ...]
    #: This item's docstring, as a sequence of lines
    docstring_lines: tuple[str, ...]
    #: The item's signature lines, for use in the directive
    signatures: tuple[str, ...] = ()

    _docstrings_has_hide_value: bool = False
    _obj: Any
    _obj___module__: str | None

    @property
    def name(self) -> str:
        """The name of the item"""
        return self.parts[-1]

    @property
    def object_name(self) -> str:
        if self._obj is RUNTIME_INSTANCE_ATTRIBUTE or self._obj is UNINITIALIZED_ATTR:
            return ''
        return self.name

    @property
    def full_name(self) -> str:
        return '.'.join((self.module_name, *self.parts))

    @property
    def parent_names(self) -> tuple[str, ...]:
        return self.parts[:-1]

    @property
    def dotted_parts(self) -> str:
        return '.'.join(self.parts)

    @property
    def _groupwise_order_key(self) -> int:
        return 0

    @property
    def canonical_module_name(self) -> str:
        if self._obj___module__ is not None:
            return self._obj___module__
        return self.module_name


@dataclasses.dataclass(frozen=False, kw_only=True, slots=True)
class _ModuleProperties(_ItemProperties):
    obj_type: Literal['module'] = 'module'
    parts: tuple[()] = ()  # modules have no parts

    file_path: Path | None
    all: Sequence[str] | None

    @property
    def name(self) -> str:
        return self.module_name.rpartition('.')[2]

    @property
    def object_name(self) -> str:
        return ''

    @property
    def full_name(self) -> str:
        return self.module_name

    @property
    def parent_names(self) -> tuple[str, ...]:
        return tuple(self.module_name.split('.')[:-1])


@dataclasses.dataclass(frozen=False, kw_only=True, slots=True)
class _ClassDefProperties(_ItemProperties):
    obj_type: Literal['class', 'exception']

    bases: Sequence[tuple[str, ...]] | None

    _obj___name__: str | None
    _obj___qualname__: str | None
    _obj_bases: tuple[str, ...]
    _obj_is_new_type: bool
    _obj_is_typevar: bool
    _signature_method_name: str = ''

    @property
    def doc_as_attr(self) -> bool:
        # if the class is documented under another name, document it
        # as data/attribute
        if self._obj___name__ is None:
            return True
        return self.parts[-1] != self._obj___name__

    @property
    def canonical_full_name(self) -> str | None:
        modname = self._obj___module__
        if modname is None:
            modname = self.module_name
        qualname = self._obj___qualname__
        if qualname is None:
            qualname = self._obj___name__
        if not modname or not qualname or '<locals>' in qualname:
            # No valid qualname found if the object is defined as locals
            return None
        return f'{modname}.{qualname}'

    @property
    def _groupwise_order_key(self) -> int:
        return 10 if self.obj_type == 'exception' else 20


@dataclasses.dataclass(frozen=False, kw_only=True, slots=True)
class _FunctionDefProperties(_ItemProperties):
    obj_type: Literal['function', 'method', 'property', 'decorator']

    properties: frozenset[_AutodocFuncProperty]

    _obj___name__: str | None
    _obj___qualname__: str | None
    _obj_property_type_annotation: str | None = None

    @property
    def is_abstractmethod(self) -> bool:
        return 'abstractmethod' in self.properties

    @property
    def is_async(self) -> bool:
        return 'async' in self.properties

    @property
    def is_classmethod(self) -> bool:
        return 'classmethod' in self.properties

    @property
    def is_final(self) -> bool:
        return 'final' in self.properties

    # @property
    # def is_singledispatch(self) -> bool:
    #     return 'singledispatch' in self.properties

    @property
    def is_staticmethod(self) -> bool:
        return 'staticmethod' in self.properties

    @property
    def _groupwise_order_key(self) -> int:
        if self.obj_type == 'method':
            if self.is_classmethod:
                # document class methods before static methods as
                # they usually behave as alternative constructors
                return 48
            if self.is_staticmethod:
                # document static members before regular methods
                return 49
            return 50
        if self.obj_type == 'property':
            return 60
        return 30


@dataclasses.dataclass(frozen=False, kw_only=True, slots=True)
class _AssignStatementProperties(_ItemProperties):
    obj_type: Literal['attribute', 'data']

    value: object
    annotation: str

    class_var: bool
    instance_var: bool

    _obj_is_generic_alias: bool
    _obj_is_attribute_descriptor: bool
    _obj_is_mock: bool
    _obj_is_sentinel: (
        RUNTIME_INSTANCE_ATTRIBUTE_T | SLOTS_ATTR_T | UNINITIALIZED_ATTR_T | None
    )
    _obj_repr_rst: str
    _obj_type_annotation: str | None

    @property
    def _groupwise_order_key(self) -> int:
        return 40 if self.obj_type == 'data' else 60


@dataclasses.dataclass(frozen=False, kw_only=True, slots=True)
class _TypeStatementProperties(_ItemProperties):
    obj_type: Literal['type']

    _obj___name__: str | None
    _obj___qualname__: str | None
    _obj___value__: str  # The aliased annotation

    @property
    def _groupwise_order_key(self) -> int:
        return 70



"""Docutils-native XML and pseudo-XML writers."""

from __future__ import annotations

from typing import TYPE_CHECKING

from docutils.writers import docutils_xml

if TYPE_CHECKING:
    from typing import Any

    from docutils.writers.docutils_xml import XMLTranslator

    from sphinx.builders import Builder


class XMLWriter(docutils_xml.Writer):
    output: str

    def __init__(self, builder: Builder) -> None:
        super().__init__()
        self.builder = builder
        self._config = builder.config

    def translate(self, *args: Any, **kwargs: Any) -> None:
        assert self.document is not None
        self.document.settings.newlines = self.document.settings.indents = (
            self._config.xml_pretty
        )
        self.document.settings.xml_declaration = True
        self.document.settings.doctype_declaration = True

        # copied from docutils.writers.docutils_xml.Writer.translate()
        # so that we can override the translator class
        visitor: XMLTranslator = self.builder.create_translator(self.document)  # type: ignore[assignment]
        self.visitor = visitor
        self.document.walkabout(visitor)
        self.output = ''.join(visitor.output)


class PseudoXMLWriter(docutils_xml.Writer):
    supported = ('pprint', 'pformat', 'pseudoxml')
    """Formats this writer supports."""

    config_section = 'pseudoxml writer'
    config_section_dependencies = ('writers',)

    output: str
    """Final translated form of `document`."""

    def __init__(self, builder: Builder) -> None:
        super().__init__()
        self.builder = builder

    def translate(self) -> None:
        assert self.document is not None
        self.output = self.document.pformat()

    def supports(self, format: str) -> bool:
        """All format-specific elements are supported."""
        return True



"""To publish HTML docs at GitHub Pages, create .nojekyll file."""

from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

import sphinx

if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment
    from sphinx.util.typing import ExtensionMetadata


def _get_domain_from_url(url: str) -> str:
    """Get the domain from a URL."""
    return url and urllib.parse.urlparse(url).hostname or ''


def create_nojekyll_and_cname(app: Sphinx, env: BuildEnvironment) -> None:
    """Manage the ``.nojekyll`` and ``CNAME`` files for GitHub Pages.

    For HTML-format builders (e.g. 'html', 'dirhtml') we unconditionally create
    the ``.nojekyll`` file to signal that GitHub Pages should not run Jekyll
    processing.

    If the :confval:`html_baseurl` option is set, we also create a CNAME file
    with the domain from ``html_baseurl``, so long as it is not a ``github.io``
    domain.

    If this extension is loaded and the domain in ``html_baseurl`` no longer
    requires a CNAME file, we remove any existing ``CNAME`` files from the
    output directory.
    """
    if app.builder.format != 'html':
        return

    app.builder.outdir.joinpath('.nojekyll').touch()
    cname_path = app.builder.outdir / 'CNAME'

    domain = _get_domain_from_url(app.config.html_baseurl)
    # Filter out GitHub Pages domains, as they do not require CNAME files.
    if domain and not domain.endswith('.github.io'):
        with open(cname_path, 'w', encoding='utf-8') as f:
            # NOTE: don't write a trailing newline. The `CNAME` file that's
            # auto-generated by the GitHub UI doesn't have one.
            f.write(domain)
    else:
        cname_path.unlink(missing_ok=True)


def setup(app: Sphinx) -> ExtensionMetadata:
    app.connect('env-updated', create_nojekyll_and_cname)
    return {
        'version': sphinx.__display_version__,
        'parallel_read_safe': True,
    }



"""Sphinx deprecation classes and utilities."""

from __future__ import annotations

import os
import warnings


class RemovedInSphinx10Warning(DeprecationWarning):
    pass


class RemovedInSphinx11Warning(PendingDeprecationWarning):
    pass


RemovedInNextVersionWarning = RemovedInSphinx10Warning


# By default, all Sphinx deprecation warnings will be emitted.
# To avoid this, set the environment variable: PYTHONWARNINGS=
if 'PYTHONWARNINGS' not in os.environ:
    warnings.filterwarnings('default', category=RemovedInNextVersionWarning)


def _deprecation_warning(
    module: str,
    attribute: str,
    canonical_name: str = '',
    *,
    remove: tuple[int, int],
    raises: bool = False,
) -> None:
    """Helper function for module-level deprecations using ``__getattr__``.

    :param module: The module containing a deprecated object.
    :param attribute: The name of the deprecated object.
    :param canonical_name: Optional fully-qualified name for its replacement.
    :param remove: Target version for removal.
    :param raises: Indicate whether to raise an exception instead of a warning.

    When *raises* is ``True``, an :exc:`AttributeError` is raised instead
    of emitting a warning so that it is easy to locate deprecated objects
    in tests that could suppress deprecation warnings.

    Usage::

       # deprecated name -> (object to return, canonical path or empty string, removal version)
       _DEPRECATED_OBJECTS = {
           'deprecated_name': (
               object_to_return,
               'fully_qualified_replacement_name',
               (10, 0),
           ),
       }


       def __getattr__(name: str) -> Any:
           if name not in _DEPRECATED_OBJECTS:
               msg = f'module {__name__!r} has no attribute {name!r}'
               raise AttributeError(msg)

           from sphinx.deprecation import _deprecation_warning

           deprecated_object, canonical_name, remove = _DEPRECATED_OBJECTS[name]
           _deprecation_warning(__name__, name, canonical_name, remove=remove)
           return deprecated_object
    """
    if remove == (10, 0):
        warning_class: type[Warning] = RemovedInSphinx10Warning
    elif remove == (11, 0):
        warning_class = RemovedInSphinx11Warning
    else:
        msg = f'removal version {remove!r} is invalid!'
        raise RuntimeError(msg)

    qualname = f'{module}.{attribute}'
    if canonical_name:
        message = (
            f'The alias {qualname!r} is deprecated, use {canonical_name!r} instead.'
        )
    else:
        message = f'{qualname!r} is deprecated.'

    if raises:
        raise AttributeError(message)

    message = f'{message} Check CHANGES for Sphinx API modifications.'
    warnings.warn(message, warning_class, stacklevel=3)



"""Builder superclass for all builders."""

from __future__ import annotations

import os
import os.path
import re
from datetime import datetime
from typing import TYPE_CHECKING

import babel.dates
from babel.messages.mofile import write_mo
from babel.messages.pofile import read_po

from sphinx.errors import SphinxError
from sphinx.locale import __
from sphinx.util import logging
from sphinx.util._pathlib import _StrPath
from sphinx.util.osutil import SEP, _last_modified_time

if TYPE_CHECKING:
    import datetime as dt
    from collections.abc import Iterator
    from typing import Protocol

    from babel.core import Locale

    from sphinx.environment import BuildEnvironment

    class DateFormatter(Protocol):
        def __call__(
            self,
            date: dt.date | None = ...,
            format: str = ...,
            locale: str | Locale | None = ...,
        ) -> str: ...

    class TimeFormatter(Protocol):
        def __call__(
            self,
            time: dt.time | dt.datetime | float | None = ...,
            format: str = ...,
            tzinfo: dt.tzinfo | None = ...,
            locale: str | Locale | None = ...,
        ) -> str: ...

    class DatetimeFormatter(Protocol):
        def __call__(
            self,
            datetime: dt.date | dt.time | float | None = ...,
            format: str = ...,
            tzinfo: dt.tzinfo | None = ...,
            locale: str | Locale | None = ...,
        ) -> str: ...

    type Formatter = DateFormatter | TimeFormatter | DatetimeFormatter

from datetime import UTC

logger = logging.getLogger(__name__)


class CatalogInfo:
    __slots__ = 'base_dir', 'domain', 'charset'

    def __init__(
        self, base_dir: str | os.PathLike[str], domain: str, charset: str
    ) -> None:
        self.base_dir = _StrPath(base_dir)
        self.domain = domain
        self.charset = charset

    @property
    def po_file(self) -> str:
        return f'{self.domain}.po'

    @property
    def mo_file(self) -> str:
        return f'{self.domain}.mo'

    @property
    def po_path(self) -> _StrPath:
        return self.base_dir / self.po_file

    @property
    def mo_path(self) -> _StrPath:
        return self.base_dir / self.mo_file

    def is_outdated(self) -> bool:
        return not self.mo_path.exists() or (
            _last_modified_time(self.mo_path) < _last_modified_time(self.po_path)
        )

    def write_mo(self, locale: str, use_fuzzy: bool = False) -> None:
        with open(self.po_path, encoding=self.charset) as file_po:
            try:
                po = read_po(file_po, locale)
            except Exception as exc:
                logger.warning(
                    __('reading error: %s, %s'),
                    self.po_path,
                    exc,
                    type='i18n',
                    subtype='not_readable',
                )
                return

        with open(self.mo_path, 'wb') as file_mo:
            try:
                write_mo(file_mo, po, use_fuzzy)
            except Exception as exc:
                logger.warning(
                    __('writing error: %s, %s'),
                    self.mo_path,
                    exc,
                    type='i18n',
                    subtype='not_writeable',
                )


class CatalogRepository:
    """A repository for message catalogs."""

    def __init__(
        self,
        basedir: str | os.PathLike[str],
        locale_dirs: list[str],
        language: str,
        encoding: str,
    ) -> None:
        self.basedir = _StrPath(basedir)
        self._locale_dirs = locale_dirs
        self.language = language
        self.encoding = encoding

    @property
    def locale_dirs(self) -> Iterator[_StrPath]:
        if not self.language:
            return

        for locale_dir in self._locale_dirs:
            locale_path = self.basedir / locale_dir / self.language / 'LC_MESSAGES'
            if locale_path.exists():
                yield self.basedir / locale_dir
            else:
                logger.verbose(__('locale_dir %s does not exist'), locale_path)

    @property
    def pofiles(self) -> Iterator[tuple[_StrPath, _StrPath]]:
        for locale_dir in self.locale_dirs:
            locale_path = locale_dir / self.language / 'LC_MESSAGES'
            for abs_path in locale_path.rglob('*.po'):
                rel_path = abs_path.relative_to(locale_path)
                # skip dot-directories
                if any(part.startswith('.') for part in rel_path.parts[:-1]):
                    continue
                yield locale_path, rel_path

    @property
    def catalogs(self) -> Iterator[CatalogInfo]:
        for basedir, filename in self.pofiles:
            domain = filename.with_suffix('').as_posix()
            yield CatalogInfo(basedir, domain, self.encoding)


def docname_to_domain(docname: str, compaction: bool | str) -> str:
    """Convert docname to domain for catalogs."""
    if isinstance(compaction, str):
        return compaction
    if compaction:
        return docname.partition(SEP)[0]
    else:
        return docname


# date_format mappings: ustrftime() to babel.dates.format_datetime()
date_format_mappings = {
    '%a':  'EEE',     # Weekday as locale's abbreviated name.
    '%A':  'EEEE',    # Weekday as locale's full name.
    '%b':  'MMM',     # Month as locale's abbreviated name.
    '%B':  'MMMM',    # Month as locale's full name.
    '%c':  'medium',  # Locale's appropriate date and time representation.
    '%-d': 'd',       # Day of the month as a decimal number.
    '%d':  'dd',      # Day of the month as a zero-padded decimal number.
    '%-H': 'H',       # Hour (24-hour clock) as a decimal number [0,23].
    '%H':  'HH',      # Hour (24-hour clock) as a zero-padded decimal number [00,23].
    '%-I': 'h',       # Hour (12-hour clock) as a decimal number [1,12].
    '%I':  'hh',      # Hour (12-hour clock) as a zero-padded decimal number [01,12].
    '%-j': 'D',       # Day of the year as a decimal number.
    '%j':  'DDD',     # Day of the year as a zero-padded decimal number.
    '%-m': 'M',       # Month as a decimal number.
    '%m':  'MM',      # Month as a zero-padded decimal number.
    '%-M': 'm',       # Minute as a decimal number [0,59].
    '%M':  'mm',      # Minute as a zero-padded decimal number [00,59].
    '%p':  'a',       # Locale's equivalent of either AM or PM.
    '%-S': 's',       # Second as a decimal number.
    '%S':  'ss',      # Second as a zero-padded decimal number.
    '%U':  'WW',      # Week number of the year (Sunday as the first day of the week)
                      # as a zero padded decimal number. All days in a new year preceding
                      # the first Sunday are considered to be in week 0.
    '%w':  'e',       # Weekday as a decimal number, where 0 is Sunday and 6 is Saturday.
    '%-W': 'W',       # Week number of the year (Monday as the first day of the week)
                      # as a decimal number. All days in a new year preceding the first
                      # Monday are considered to be in week 0.
    '%W':  'WW',      # Week number of the year (Monday as the first day of the week)
                      # as a zero-padded decimal number.
    '%x':  'medium',  # Locale's appropriate date representation.
    '%X':  'medium',  # Locale's appropriate time representation.
    '%y':  'YY',      # Year without century as a zero-padded decimal number.
    '%Y':  'yyyy',    # Year with century as a decimal number.
    '%Z':  'zzz',     # Time zone name (no characters if no time zone exists).
    '%z':  'ZZZ',     # UTC offset in the form ±HHMM[SS[.ffffff]]
                      # (empty string if the object is naive).
    '%%':  '%',
}  # fmt: skip

date_format_re = re.compile('(%s)' % '|'.join(date_format_mappings))


def babel_format_date(
    date: dt.datetime,
    format: str,
    locale: str,
    formatter: Formatter = babel.dates.format_date,
) -> str:
    # Check if we have the tzinfo attribute. If not we cannot do any time
    # related formats.
    if not hasattr(date, 'tzinfo'):
        formatter = babel.dates.format_date

    if not locale:
        # Babel would not accept a falsy locale
        # (or would try to fall back to the LC_TIME
        # locale, which would be not what was requested),
        # so we can just short-cut to English, as we
        # would for the `"fallback to English"` case.
        locale = 'en'

    try:
        return formatter(date, format, locale=locale)
    except (ValueError, babel.core.UnknownLocaleError):
        # fallback to English
        logger.warning(
            __('Invalid Babel locale: %r.'),
            locale,
            type='i18n',
            subtype='babel',
        )
        return formatter(date, format, locale='en')
    except AttributeError:
        logger.warning(
            __(
                'Invalid date format. Quote the string by single quote '
                'if you want to output it directly: %s'
            ),
            format,
            type='i18n',
            subtype='babel',
        )
        return format


def format_date(
    format: str,
    *,
    date: datetime | None = None,
    language: str,
    local_time: bool = False,
) -> str:
    if date is None:
        # If time is not specified, try to use $SOURCE_DATE_EPOCH variable
        # See https://wiki.debian.org/ReproducibleBuilds/TimestampsProposal
        source_date_epoch = os.getenv('SOURCE_DATE_EPOCH')
        if source_date_epoch is not None:
            date = datetime.fromtimestamp(float(source_date_epoch), tz=UTC)
            # If SOURCE_DATE_EPOCH is set, users likely want a reproducible result,
            # so enforce GMT/UTC for consistency.
            local_time = False
        else:
            date = datetime.now(tz=UTC)

    if local_time:
        # > If called with tz=None, the system local time zone
        # > is assumed for the target time zone.
        # https://docs.python.org/dev/library/datetime.html#datetime.datetime.astimezone
        date = date.astimezone(tz=None)

    result = []
    tokens = date_format_re.split(format)
    for token in tokens:
        if token in date_format_mappings:
            babel_format = date_format_mappings.get(token, '')

            # Check if we have to use a different babel formatter then
            # format_datetime, because we only want to format a date
            # or a time.
            function: Formatter
            if token == '%x':
                function = babel.dates.format_date
            elif token == '%X':
                function = babel.dates.format_time
            else:
                function = babel.dates.format_datetime

            result.append(
                babel_format_date(
                    date, babel_format, locale=language, formatter=function
                )
            )
        else:
            result.append(token)

    return ''.join(result)


def get_image_filename_for_language(
    filename: str | os.PathLike[str],
    env: BuildEnvironment,
) -> str:
    root, ext = os.path.splitext(filename)
    dirname = os.path.dirname(root)
    docpath = os.path.dirname(env.current_document.docname)
    try:
        return env.config.figure_language_filename.format(
            root=root,
            ext=ext,
            path=dirname and dirname + SEP,
            basename=os.path.basename(root),
            docpath=docpath and docpath + SEP,
            language=env.config.language,
        )
    except KeyError as exc:
        msg = f'Invalid figure_language_filename: {exc!r}'
        raise SphinxError(msg) from exc


def search_image_for_language(filename: str, env: BuildEnvironment) -> str:
    translated = get_image_filename_for_language(filename, env)
    _, abspath = env.relfn2path(translated)
    if os.path.exists(abspath):
        return translated
    else:
        return filename



"""Image converter extension for Sphinx"""

from __future__ import annotations

import subprocess
import sys
from subprocess import CalledProcessError
from typing import TYPE_CHECKING

import sphinx
from sphinx.errors import ExtensionError
from sphinx.locale import __
from sphinx.transforms.post_transforms.images import ImageConverter
from sphinx.util import logging

if TYPE_CHECKING:
    import os

    from sphinx.application import Sphinx
    from sphinx.util.typing import ExtensionMetadata

logger = logging.getLogger(__name__)


class ImagemagickConverter(ImageConverter):
    conversion_rules = [
        ('image/svg+xml', 'image/png'),
        ('image/gif', 'image/png'),
        ('application/pdf', 'image/png'),
        ('application/illustrator', 'image/png'),
        ('image/webp', 'image/png'),
    ]

    def is_available(self) -> bool:
        """Confirms the converter is available or not."""
        try:
            args = [self.config.image_converter, '-version']
            logger.debug('Invoking %r ...', args)
            subprocess.run(args, capture_output=True, check=True)
            return True
        except OSError as exc:
            logger.warning(
                __(
                    'Unable to run the image conversion command %r. '
                    "'sphinx.ext.imgconverter' requires ImageMagick by default. "
                    "Ensure it is installed, or set the 'image_converter' option "
                    'to a custom conversion command.\n\n'
                    'Traceback: %s',
                ),
                self.config.image_converter,
                exc,
            )
            return False
        except CalledProcessError as exc:
            logger.warning(
                __('convert exited with error:\n[stderr]\n%r\n[stdout]\n%r'),
                exc.stderr,
                exc.stdout,
            )
            return False

    def convert(
        self, _from: str | os.PathLike[str], _to: str | os.PathLike[str]
    ) -> bool:
        """Converts the image to expected one."""
        try:
            # append an index 0 to source filename to pick up the first frame
            # (or first page) of image (ex. Animation GIF, PDF)
            from_ = f'{_from}[0]'

            args = [
                self.config.image_converter,
                *self.config.image_converter_args,
                from_,
                _to,
            ]
            logger.debug('Invoking %r ...', args)
            subprocess.run(args, capture_output=True, check=True)
            return True
        except OSError:
            logger.warning(
                __(
                    'convert command %r cannot be run, check the image_converter setting'
                ),
                self.config.image_converter,
            )
            return False
        except CalledProcessError as exc:
            raise ExtensionError(
                __('convert exited with error:\n[stderr]\n%r\n[stdout]\n%r')
                % (exc.stderr, exc.stdout)
            ) from exc


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_post_transform(ImagemagickConverter)
    if sys.platform == 'win32':
        # On Windows, we use Imagemagik v7 by default to avoid the trouble for
        # convert.exe bundled with Windows.
        app.add_config_value('image_converter', 'magick', 'env', types=frozenset({str}))
        app.add_config_value(
            'image_converter_args', ['convert'], 'env', types=frozenset({list, tuple})
        )
    else:
        # On other platform, we use Imagemagick v6 by default.  Especially,
        # Debian/Ubuntu are still based of v6.  So we can't use "magick" command
        # for these platforms.
        app.add_config_value(
            'image_converter', 'convert', 'env', types=frozenset({str})
        )
        app.add_config_value(
            'image_converter_args', [], 'env', types=frozenset({list, tuple})
        )

    return {
        'version': sphinx.__display_version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Global creation environment."""

from __future__ import annotations

import functools
import os
import pickle
import warnings
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING

from sphinx import addnodes
from sphinx.deprecation import RemovedInSphinx11Warning, _deprecation_warning
from sphinx.domains._domains_container import _DomainsContainer
from sphinx.environment.adapters import toctree as toctree_adapters
from sphinx.errors import (
    BuildEnvironmentError,
    DocumentError,
    ExtensionError,
    SphinxError,
)
from sphinx.locale import __
from sphinx.transforms import SphinxTransformer
from sphinx.util import logging
from sphinx.util._files import DownloadFiles, FilenameUniqDict
from sphinx.util._pathlib import _StrPathProperty
from sphinx.util._serialise import stable_str
from sphinx.util._timestamps import _format_rfc3339_microseconds
from sphinx.util.docutils import LoggingReporter
from sphinx.util.i18n import CatalogRepository, docname_to_domain
from sphinx.util.nodes import is_translatable
from sphinx.util.osutil import _last_modified_time, _relative_path

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator, Mapping, Set
    from typing import Any, Final, Literal

    from docutils import nodes
    from docutils.nodes import Node
    from docutils.parsers import Parser

    from sphinx.application import Sphinx
    from sphinx.builders import Builder
    from sphinx.config import Config
    from sphinx.domains import Domain
    from sphinx.domains.c._symbol import Symbol as CSymbol
    from sphinx.domains.cpp._symbol import Symbol as CPPSymbol
    from sphinx.events import EventManager
    from sphinx.extension import Extension
    from sphinx.project import Project
    from sphinx.registry import SphinxComponentRegistry
    from sphinx.util._pathlib import _StrPath
    from sphinx.util.tags import Tags

logger = logging.getLogger(__name__)

default_settings: dict[str, Any] = {
    'auto_id_prefix': 'id',
    'image_loading': 'link',
    'embed_stylesheet': False,
    'cloak_email_addresses': True,
    'pep_base_url': 'https://peps.python.org/',
    'pep_references': None,
    'rfc_base_url': 'https://datatracker.ietf.org/doc/html/',
    'rfc_references': None,
    'input_encoding': 'utf-8-sig',
    'doctitle_xform': False,
    'sectsubtitle_xform': False,
    'section_self_link': False,
    'halt_level': 5,
    'file_insertion_enabled': True,
    'smartquotes_locales': [],
}

# This is increased every time an environment attribute is added
# or changed to properly invalidate pickle files.
ENV_VERSION = 66

# config status
CONFIG_UNSET = -1
CONFIG_OK = 1
CONFIG_NEW = 2
CONFIG_CHANGED = 3
CONFIG_EXTENSIONS_CHANGED = 4

CONFIG_CHANGED_REASON = {
    CONFIG_NEW: __('new config'),
    CONFIG_CHANGED: __('config changed'),
    CONFIG_EXTENSIONS_CHANGED: __('extensions changed'),
}


versioning_conditions: dict[str, Literal[False] | Callable[[Node], bool]] = {
    'none': False,
    'text': is_translatable,
}


class BuildEnvironment:
    """The environment in which the ReST files are translated.
    Stores an inventory of cross-file targets and provides doctree
    transformations to resolve links to them.
    """

    # --------- ENVIRONMENT INITIALIZATION -------------------------------------

    srcdir = _StrPathProperty()
    doctreedir = _StrPathProperty()

    # builder is created after the environment.
    _builder_cls: type[Builder]

    def __init__(self, app: Sphinx) -> None:
        self._app: Sphinx = app
        self.doctreedir = app.doctreedir
        self.srcdir = app.srcdir
        self.config: Config = None  # type: ignore[assignment]
        self.config_status: int = CONFIG_UNSET
        self.config_status_extra: str = ''
        self.events: EventManager = app.events
        self.project: Project = app.project
        self.version: Mapping[str, int] = _get_env_version(app.extensions)

        # the method of doctree versioning; see set_versioning_method
        self.versioning_condition: Literal[False] | Callable[[Node], bool] | None = None
        self.versioning_compare: bool | None = None

        # the docutils settings for building
        self.settings: dict[str, Any] = default_settings.copy()
        self.settings['env'] = self

        # All "docnames" here are /-separated and relative and exclude
        # the source suffix.

        # docname -> time of reading (in integer microseconds)
        # contains all read docnames
        self.all_docs: dict[str, int] = {}
        # docname -> set of dependent file
        # names, relative to documentation root
        self.dependencies: dict[str, set[_StrPath]] = defaultdict(set)
        # docname -> set of included file
        # docnames included from other documents
        self.included: dict[str, set[str]] = defaultdict(set)
        # docnames to re-read unconditionally on next build
        self.reread_always: set[str] = set()

        self._pickled_doctree_cache: dict[str, bytes] = {}
        """In-memory cache for reading pickled doctrees from disk.
        docname -> pickled doctree

        This cache is used in the ``get_doctree`` method to avoid reading the
        doctree from disk multiple times.
        """

        self._write_doc_doctree_cache: dict[str, nodes.document] = {}
        """In-memory cache for unpickling doctrees from disk.
        docname -> doctree

        Items are added in ``Builder.write_doctree``, during the read phase,
        then used only in the ``get_and_resolve_doctree`` method.
        """

        # File metadata
        # docname -> dict of metadata items
        self.metadata: dict[str, dict[str, Any]] = defaultdict(dict)

        # TOC inventory
        # docname -> title node
        self.titles: dict[str, nodes.title] = {}
        # docname -> title node; only different if
        # set differently with title directive
        self.longtitles: dict[str, nodes.title] = {}
        # docname -> table of contents nodetree
        self.tocs: dict[str, nodes.bullet_list] = {}
        # docname -> number of real entries
        self.toc_num_entries: dict[str, int] = {}

        # used to determine when to show the TOC
        # in a sidebar (don't show if it's only one item)
        # docname -> dict of sectionid -> number
        self.toc_secnumbers: dict[str, dict[str, tuple[int, ...]]] = {}
        # docname -> dict of figtype -> dict of figureid -> number
        self.toc_fignumbers: dict[str, dict[str, dict[str, tuple[int, ...]]]] = {}

        # docname -> list of toctree includefiles
        self.toctree_includes: dict[str, list[str]] = {}
        # docname -> set of files (containing its TOCs) to rebuild too
        self.files_to_rebuild: dict[str, set[str]] = {}
        # docnames that have :glob: toctrees
        self.glob_toctrees: set[str] = set()
        # docnames that have :numbered: toctrees
        self.numbered_toctrees: set[str] = set()

        # domain-specific inventories, here to be pickled
        # domainname -> domain-specific dict
        self.domaindata: dict[str, dict[str, Any]] = {}

        # these map absolute path -> (docnames, unique filename)
        self.images: FilenameUniqDict = FilenameUniqDict()
        # filename -> (set of docnames, destination)
        self.dlfiles: DownloadFiles = DownloadFiles()

        # the original URI for images
        self.original_image_uri: dict[_StrPath, str] = {}

        # temporary data storage while reading a document
        self.current_document: _CurrentDocument = _CurrentDocument()
        # context for cross-references (e.g. current module or class)
        # this is similar to ``self.current_document``,
        # but will for example be copied to attributes of "any" cross references
        self.ref_context: dict[str, Any] = {}

        # search index data

        # docname -> title
        self._search_index_titles: dict[str, str | None] = {}
        # docname -> filename
        self._search_index_filenames: dict[str, str] = {}
        # stemmed words -> set(docname)
        self._search_index_mapping: dict[str, set[str]] = {}
        # stemmed words in titles -> set(docname)
        self._search_index_title_mapping: dict[str, set[str]] = {}
        # docname -> all titles in document
        self._search_index_all_titles: dict[str, list[tuple[str, str | None]]] = {}
        # docname -> list(index entry)
        self._search_index_index_entries: dict[str, list[tuple[str, str, str]]] = {}
        # objtype -> index
        self._search_index_objtypes: dict[tuple[str, str], int] = {}
        # objtype index -> (domain, type, objname (localized))
        self._search_index_objnames: dict[int, tuple[str, str, str]] = {}

        # all the registered domains, set by the application
        self.domains: _DomainsContainer = _DomainsContainer._from_environment(
            self, registry=app.registry
        )

        # set up environment
        self.setup(app)

    def __getstate__(self) -> dict[str, Any]:
        """Obtains serializable data for pickling."""
        __dict__ = self.__dict__.copy()
        # clear unpickleable attributes
        __dict__.update(_app=None, domains=None, events=None)
        # clear in-memory doctree caches, to reduce memory consumption and
        # ensure that, upon restoring the state, the most recent pickled files
        # on the disk are used instead of those from a possibly outdated state
        __dict__.update(_pickled_doctree_cache={}, _write_doc_doctree_cache={})
        # clear attributes set in Sphinx._post_init_env()
        __dict__.pop('_builder_cls', None)
        return __dict__

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__.update(state)

    def setup(self, app: Sphinx) -> None:
        """Set up BuildEnvironment object."""
        if self.version and self.version != _get_env_version(app.extensions):
            raise BuildEnvironmentError(__('build environment version not current'))
        if self.srcdir and self.srcdir != app.srcdir:
            raise BuildEnvironmentError(__('source directory has changed'))

        if self.project:
            app.project.restore(self.project)

        self._app = app
        self.doctreedir = app.doctreedir
        self.events = app.events
        self.srcdir = app.srcdir
        self.project = app.project
        self.version = _get_env_version(app.extensions)

        # initialise domains
        if self.domains is None:
            # if we are unpickling an environment, we need to recreate the domains
            self.domains = _DomainsContainer._from_environment(
                self, registry=app.registry
            )
        # setup domains (must do after all initialization)
        self.domains._setup()

        # Initialise config.
        # The old config is self.config, restored from the pickled environment.
        # The new config is app.config, always recreated from ``conf.py``
        self.config_status, self.config_status_extra = self._config_status(
            old_config=self.config,
            new_config=app.config,
            verbosity=app.config.verbosity,
        )
        self.config = app.config

        # initialize settings
        self._update_settings(app.config)

    @property
    def app(self) -> Sphinx:
        _deprecation_warning(__name__, 'BuildEnvironment.app', remove=(11, 0))
        return self._app

    @app.setter
    def app(self, app: Sphinx) -> None:
        _deprecation_warning(__name__, 'BuildEnvironment.app', remove=(11, 0))
        self._app = app

    @app.deleter
    def app(self) -> None:
        _deprecation_warning(__name__, 'BuildEnvironment.app', remove=(11, 0))
        del self._app

    @property
    def _registry(self) -> SphinxComponentRegistry:
        return self._app.registry

    @property
    def _tags(self) -> Tags:
        return self._app.tags

    @staticmethod
    def _config_status(
        *, old_config: Config | None, new_config: Config, verbosity: int
    ) -> tuple[int, str]:
        """Report the differences between two Config objects.

        Returns a triple of:

        1. The new configuration
        2. A status code indicating how the configuration has changed.
        3. A status message indicating what has changed.
        """
        if old_config is None:
            return CONFIG_NEW, ''

        if old_config.extensions != new_config.extensions:
            old_extensions = set(old_config.extensions)
            new_extensions = set(new_config.extensions)
            extensions = old_extensions ^ new_extensions
            if len(extensions) == 1:
                extension = extensions.pop()
            else:
                extension = f'{len(extensions)}'
            return CONFIG_EXTENSIONS_CHANGED, f' ({extension!r})'

        # Log any changes in configuration keys
        if changed_keys := _differing_config_keys(old_config, new_config):
            changed_num = len(changed_keys)
            if changed_num == 1:
                logger.info(
                    __('The configuration has changed (1 option: %r)'),
                    next(iter(changed_keys)),
                )
            elif changed_num <= 5 or verbosity >= 1:
                logger.info(
                    __('The configuration has changed (%d options: %s)'),
                    changed_num,
                    ', '.join(map(repr, sorted(changed_keys))),
                )
            else:
                logger.info(
                    __('The configuration has changed (%d options: %s, ...)'),
                    changed_num,
                    ', '.join(map(repr, sorted(changed_keys)[:5])),
                )

        # check if a config value was changed that affects how doctrees are read
        for item in new_config.filter(frozenset({'env'})):
            if old_config[item.name] != item.value:
                return CONFIG_CHANGED, f' ({item.name!r})'

        return CONFIG_OK, ''

    def _update_settings(self, config: Config) -> None:
        """Update settings by new config."""
        self.settings['input_encoding'] = config.source_encoding
        self.settings['trim_footnote_reference_space'] = (
            config.trim_footnote_reference_space
        )
        self.settings['language_code'] = config.language

        # Allow to disable by 3rd party extension (workaround)
        self.settings.setdefault('smart_quotes', True)

    def set_versioning_method(
        self, method: str | Callable[[Node], bool], compare: bool
    ) -> None:
        """Set the doctree versioning method for this environment.

        Versioning methods are a builder property; only builders with the same
        versioning method can share the same doctree directory.  Therefore, we
        raise an exception if the user tries to use an environment with an
        incompatible versioning method.
        """
        condition: Literal[False] | Callable[[Node], bool]
        if callable(method):
            condition = method
        else:
            if method not in versioning_conditions:
                raise ValueError('invalid versioning method: %r' % method)
            condition = versioning_conditions[method]

        if self.versioning_condition not in {None, condition}:
            msg = __(
                'This environment is incompatible with the '
                'selected builder, please choose another '
                'doctree directory.'
            )
            raise SphinxError(msg)
        self.versioning_condition = condition
        self.versioning_compare = compare

    def clear_doc(self, docname: str) -> None:
        """Remove all traces of a source file in the inventory."""
        if docname in self.all_docs:
            self.all_docs.pop(docname, None)
            self.included.pop(docname, None)
            self.reread_always.discard(docname)

        self.domains._clear_doc(docname)

    def merge_info_from(
        self, docnames: Iterable[str], other: BuildEnvironment, app: Sphinx
    ) -> None:
        """Merge global information gathered about *docnames* while reading them
        from the *other* environment.

        This possibly comes from a parallel build process.
        """
        docnames = frozenset(docnames)
        for docname in docnames:
            self.all_docs[docname] = other.all_docs[docname]
            self.included[docname] = other.included[docname]
            if docname in other.reread_always:
                self.reread_always.add(docname)

        self.domains._merge_domain_data(docnames, other.domaindata)
        self.events.emit('env-merge-info', self, docnames, other)

    def path2doc(self, filename: str | os.PathLike[str]) -> str | None:
        """Return the docname for the filename if the file is document.

        *filename* should be absolute or relative to the source directory.
        """
        return self.project.path2doc(filename)

    def doc2path(self, docname: str, base: bool = True) -> _StrPath:
        """Return the filename for the document name.

        If *base* is True, return absolute path under self.srcdir.
        If *base* is False, return relative path to self.srcdir.
        """
        return self.project.doc2path(docname, absolute=base)

    def relfn2path(
        self, filename: str | Path, docname: str | None = None
    ) -> tuple[str, str]:
        """Return paths to a file referenced from a document, relative to
        documentation root and absolute.

        In the input "filename", absolute filenames are taken as relative to the
        source dir, while relative filenames are relative to the dir of the
        containing document.
        """
        file_name = Path(filename)
        if file_name.parts[:1] in {('/',), ('\\',)}:
            abs_fn = self.srcdir.joinpath(*file_name.parts[1:]).resolve()
        else:
            if not docname:
                if self.docname:
                    docname = self.docname
                else:
                    msg = 'docname'
                    raise KeyError(msg)
            doc_dir = self.doc2path(docname, base=False).parent
            abs_fn = self.srcdir.joinpath(doc_dir, file_name).resolve()

        rel_fn = _relative_path(abs_fn, self.srcdir)
        return rel_fn.as_posix(), os.fspath(abs_fn)

    @property
    def found_docs(self) -> set[str]:
        """Contains all existing docnames."""
        return self.project.docnames

    def find_files(self, config: Config, builder: Builder) -> None:
        """Find all source files in the source dir and put them in
        self.found_docs.
        """
        try:
            exclude_paths = (
                self.config.exclude_patterns
                + self.config.templates_path
                + builder.get_asset_paths()
            )
            self.project.discover(exclude_paths, self.config.include_patterns)

            # Current implementation is applying translated messages in the reading
            # phase.Therefore, in order to apply the updated message catalog, it is
            # necessary to re-process from the reading phase. Here, if dependency
            # is set for the doc source and the mo file, it is processed again from
            # the reading phase when mo is updated. In the future, we would like to
            # move i18n process into the writing phase, and remove these lines.
            if builder.use_message_catalog:
                # add catalog mo file dependency
                repo = CatalogRepository(
                    self.srcdir,
                    self.config.locale_dirs,
                    self.config.language,
                    self.config.source_encoding,
                )
                mo_paths = {c.domain: c.mo_path for c in repo.catalogs}
                for docname in self.found_docs:
                    domain = docname_to_domain(docname, self.config.gettext_compact)
                    if domain in mo_paths:
                        self.note_dependency(mo_paths[domain], docname=docname)
        except OSError as exc:
            raise DocumentError(
                __('Failed to scan documents in %s: %r') % (self.srcdir, exc)
            ) from exc

    def get_outdated_files(
        self, config_changed: bool
    ) -> tuple[set[str], set[str], set[str]]:
        """Return (added, changed, removed) sets."""
        # clear all files no longer present
        removed = self.all_docs.keys() - self.found_docs

        added: set[str] = set()
        changed: set[str] = set()

        if config_changed:
            # config values affect e.g. substitutions
            added = self.found_docs
            return added, changed, removed

        for docname in self.found_docs:
            if docname not in self.all_docs:
                logger.debug('[build target] added %r', docname)
                added.add(docname)
                continue

            # if the document has changed, rebuild
            if _has_doc_changed(
                docname,
                filename=self.doc2path(docname),
                reread_always=self.reread_always,
                doctreedir=self.doctreedir,
                all_docs=self.all_docs,
                dependencies=self.dependencies,
            ):
                changed.add(docname)
                continue

        return added, changed, removed

    def check_dependents(self, app: Sphinx, already: set[str]) -> Iterator[str]:
        to_rewrite: list[str] = []
        for docnames in self.events.emit('env-get-updated', self):
            to_rewrite.extend(docnames)
        for docname in set(to_rewrite):
            if docname not in already:
                yield docname

    # --------- SINGLE FILE READING --------------------------------------------

    def prepare_settings(self, docname: str) -> None:
        """Prepare to set up environment for reading."""
        self.current_document = _CurrentDocument(
            docname=docname,
            # defaults to the global default, but can be re-set in a document
            default_role=self.config.default_role,
            default_domain=self.domains.get(self.config.primary_domain),
        )

    # utilities to use while reading a document

    @property
    def temp_data(self) -> _CurrentDocument:
        """Returns the temporary data storage for the current document.

        Kept for backwards compatibility.
        """
        return self.current_document

    @property
    def docname(self) -> str:
        """Returns the docname of the document currently being parsed."""
        return self.current_document.docname

    @property
    def parser(self) -> Parser:
        """Returns the parser being used for to parse the current document."""
        if (parser := self.current_document._parser) is not None:
            return parser
        msg = 'parser'
        raise KeyError(msg)

    def new_serialno(self, category: str = '') -> int:
        """Return a serial number, e.g. for index entry targets.

        The number is guaranteed to be unique in the current document.
        """
        return self.current_document.new_serial_number(category)

    def note_dependency(
        self, filename: str | os.PathLike[str], *, docname: str | None = None
    ) -> None:
        """Add *filename* as a dependency of the current document.

        This means that the document will be rebuilt if this file changes.

        *filename* should be absolute or relative to the source directory.
        """
        if docname is None:
            docname = self.docname
        # this will do the right thing when *filename* is absolute too
        filename = self.srcdir / filename
        self.dependencies.setdefault(docname, set()).add(filename)

    def note_included(self, filename: str | os.PathLike[str]) -> None:
        """Add *filename* as a included from other document.

        This means the document is not orphaned.

        *filename* should be absolute or relative to the source directory.
        """
        doc = self.path2doc(filename)
        if doc:
            self.included.setdefault(self.docname, set()).add(doc)

    def note_reread(self) -> None:
        """Add the current document to the list of documents that will
        automatically be re-read at the next build.
        """
        self.reread_always.add(self.docname)

    def get_domain(self, domainname: str) -> Domain:
        """Return the domain instance with the specified name.

        Raises an ExtensionError if the domain is not registered.
        """
        try:
            return self.domains[domainname]
        except KeyError as exc:
            msg = __('Domain %r is not registered') % domainname
            raise ExtensionError(msg) from exc

    # --------- RESOLVING REFERENCES AND TOCTREES ------------------------------

    def get_doctree(self, docname: str) -> nodes.document:
        """Read the doctree for a file from the pickle and return it."""
        try:
            serialised = self._pickled_doctree_cache[docname]
        except KeyError:
            filename = self.doctreedir / f'{docname}.doctree'
            with open(filename, 'rb') as f:
                serialised = self._pickled_doctree_cache[docname] = f.read()

        doctree = pickle.loads(serialised)
        doctree.settings.env = self
        doctree.reporter = LoggingReporter(str(self.doc2path(docname)))
        return doctree

    @functools.cached_property
    def master_doctree(self) -> nodes.document:
        return self.get_doctree(self.config.root_doc)

    def get_and_resolve_doctree(
        self,
        docname: str,
        builder: Builder,
        *,
        tags: Tags = ...,  # type: ignore[assignment]
        doctree: nodes.document | None = None,
        prune_toctrees: bool = True,
        includehidden: bool = False,
    ) -> nodes.document:
        """Read the doctree from the pickle, resolve cross-references and
        toctrees and return it.
        """
        if tags is ...:
            warnings.warn(
                "'tags' will become a required keyword argument "
                'for global_toctree_for_doc() in Sphinx 11.0.',
                RemovedInSphinx11Warning,
                stacklevel=2,
            )
            tags = builder.tags

        if doctree is None:
            try:
                doctree = self._write_doc_doctree_cache.pop(docname)
                doctree.settings.env = self
                doctree.reporter = LoggingReporter(str(self.doc2path(docname)))
            except KeyError:
                doctree = self.get_doctree(docname)

        # resolve all pending cross-references
        self.apply_post_transforms(doctree, docname)

        # now, resolve all toctree nodes
        for toctreenode in doctree.findall(addnodes.toctree):
            result = toctree_adapters._resolve_toctree(
                self,
                docname,
                builder,
                toctreenode,
                prune=prune_toctrees,
                includehidden=includehidden,
                tags=tags,
            )
            if result is None:
                toctreenode.parent.replace(toctreenode, [])
            else:
                toctreenode.replace_self(result)

        return doctree

    def resolve_toctree(
        self,
        docname: str,
        builder: Builder,
        toctree: addnodes.toctree,
        prune: bool = True,
        maxdepth: int = 0,
        titles_only: bool = False,
        collapse: bool = False,
        includehidden: bool = False,
    ) -> Node | None:
        """Resolve a *toctree* node into individual bullet lists with titles
        as items, returning None (if no containing titles are found) or
        a new node.

        If *prune* is True, the tree is pruned to *maxdepth*, or if that is 0,
        to the value of the *maxdepth* option on the *toctree* node.
        If *titles_only* is True, only toplevel document titles will be in the
        resulting tree.
        If *collapse* is True, all branches not containing docname will
        be collapsed.
        """
        return toctree_adapters._resolve_toctree(
            self,
            docname,
            builder,
            toctree,
            prune=prune,
            maxdepth=maxdepth,
            titles_only=titles_only,
            collapse=collapse,
            includehidden=includehidden,
            tags=self._tags,
        )

    def resolve_references(
        self, doctree: nodes.document, fromdocname: str, builder: Builder
    ) -> None:
        self.apply_post_transforms(doctree, fromdocname)

    def apply_post_transforms(self, doctree: nodes.document, docname: str) -> None:
        """Apply all post-transforms."""
        backup = self.current_document
        new = deepcopy(backup)
        new.docname = docname
        try:
            # set env.current_document.docname during applying post-transforms
            self.current_document = new

            transformer = SphinxTransformer(doctree)
            transformer.set_environment(self)
            transformer.add_transforms(self._registry.get_post_transforms())
            transformer.apply_transforms()
        finally:
            self.current_document = backup

        # allow custom references to be resolved
        self.events.emit('doctree-resolved', doctree, docname)

    def collect_relations(self) -> dict[str, list[str | None]]:
        traversed: set[str] = set()

        relations = {}
        docnames = _traverse_toctree(
            traversed, None, self.config.root_doc, self.toctree_includes
        )
        prev_doc = None
        parent, docname = next(docnames)
        for next_parent, next_doc in docnames:
            relations[docname] = [parent, prev_doc, next_doc]
            prev_doc = docname
            docname = next_doc
            parent = next_parent

        relations[docname] = [parent, prev_doc, None]

        return relations

    def check_consistency(self) -> None:
        """Do consistency checks."""
        included = set().union(*self.included.values())
        for docname in sorted(self.all_docs):
            if docname not in self.files_to_rebuild:
                if docname == self.config.root_doc:
                    # the master file is not included anywhere ;)
                    continue
                if docname in included:
                    # the document is included from other documents
                    continue
                if 'orphan' in self.metadata[docname]:
                    continue
                logger.warning(
                    __("document isn't included in any toctree"),
                    location=docname,
                    type='toc',
                    subtype='not_included',
                )
        # Call _check_toc_parents here rather than in  _get_toctree_ancestors()
        # because that method is called multiple times per document and would
        # lead to duplicate warnings.
        _check_toc_parents(self.toctree_includes)

        # call check-consistency for all extensions
        self.domains._check_consistency()
        self.events.emit('env-check-consistency', self)


def _get_env_version(extensions: Mapping[str, Extension]) -> Mapping[str, int]:
    env_version = {
        ext.name: ext_env_version
        for ext in extensions.values()
        if (ext_env_version := ext.metadata.get('env_version'))
    }
    env_version['sphinx'] = ENV_VERSION
    return env_version


def _differing_config_keys(old: Config, new: Config) -> frozenset[str]:
    """Return a set of keys that differ between two config objects."""
    old_vals = {c.name: c.value for c in old}
    new_vals = {c.name: c.value for c in new}
    not_in_both = old_vals.keys() ^ new_vals.keys()
    different_values = {
        key
        for key in old_vals.keys() & new_vals.keys()
        if stable_str(old_vals[key]) != stable_str(new_vals[key])
    }
    return frozenset(not_in_both | different_values)


def _has_doc_changed(
    docname: str,
    *,
    filename: Path,
    reread_always: Set[str],
    doctreedir: Path,
    all_docs: Mapping[str, int],
    dependencies: Mapping[str, Set[Path]],
) -> bool:
    # check the "reread always" list
    if docname in reread_always:
        logger.debug('[build target] changed %r: re-read forced', docname)
        return True

    # if the doctree file is not there, rebuild
    doctree_path = doctreedir / f'{docname}.doctree'
    if not doctree_path.is_file():
        logger.debug('[build target] changed %r: doctree file does not exist', docname)
        return True

    # check the mtime of the document
    mtime = all_docs[docname]
    new_mtime = _last_modified_time(filename)
    if new_mtime > mtime:
        logger.debug(
            '[build target] changed: %r is outdated (%s -> %s)',
            docname,
            _format_rfc3339_microseconds(mtime),
            _format_rfc3339_microseconds(new_mtime),
        )
        return True

    # finally, check the mtime of dependencies
    if docname not in dependencies:
        return False
    for dep_path in dependencies[docname]:
        try:
            dep_path_is_file = dep_path.is_file()
        except OSError:
            return True  # give it another chance
        if not dep_path_is_file:
            logger.debug(
                '[build target] changed: %r is missing dependency %r',
                docname,
                dep_path,
            )
            return True

        try:
            dep_mtime = _last_modified_time(dep_path)
        except OSError:
            return True  # give it another chance
        if dep_mtime > mtime:
            logger.debug(
                '[build target] changed: %r is outdated due to dependency %r (%s -> %s)',
                docname,
                dep_path,
                _format_rfc3339_microseconds(mtime),
                _format_rfc3339_microseconds(dep_mtime),
            )
            return True

    return False


def _traverse_toctree(
    traversed: set[str],
    parent: str | None,
    docname: str,
    toctree_includes: dict[str, list[str]],
) -> Iterator[tuple[str | None, str]]:
    if parent == docname:
        logger.warning(
            __('self referenced toctree found. Ignored.'),
            location=docname,
            type='toc',
            subtype='circular',
        )
        return

    # traverse toctree by pre-order
    yield parent, docname
    traversed.add(docname)

    for child in toctree_includes.get(docname, ()):
        for sub_parent, sub_docname in _traverse_toctree(
            traversed, docname, child, toctree_includes
        ):
            if sub_docname not in traversed:
                yield sub_parent, sub_docname
                traversed.add(sub_docname)


def _check_toc_parents(toctree_includes: dict[str, list[str]]) -> None:
    toc_parents: dict[str, list[str]] = {}
    for parent, children in toctree_includes.items():
        for child in children:
            toc_parents.setdefault(child, []).append(parent)

    for doc, parents in sorted(toc_parents.items()):
        if len(parents) > 1:
            logger.info(
                __(
                    'document is referenced in multiple toctrees: %s, selecting: %s <- %s'
                ),
                parents,
                max(parents),
                doc,
                location=doc,
                type='toc',
                subtype='multiple_toc_parents',
            )


class _CurrentDocument:
    """Temporary data storage while reading a document.

    This class is only for internal use. Please don't use this in your extensions.
    It will be removed or changed without notice.
    The only stable API is via ``env.current_document``.
    """

    __slots__ = (
        '_parser',
        '_serial_numbers',
        '_extension_data',
        'autodoc_annotations',
        'autodoc_class',
        'autodoc_module',
        'c_last_symbol',
        'c_namespace_stack',
        'c_parent_symbol',
        'cpp_domain_name',
        'cpp_last_symbol',
        'cpp_namespace_stack',
        'cpp_parent_symbol',
        'default_domain',
        'default_role',
        'docname',
        'highlight_language',
        'obj_desc_name',
        'reading_started_at',
    )

    # Map of old-style temp_data keys to _CurrentDocument attributes
    __attr_map: Final = {
        '_parser': '_parser',
        'annotations': 'autodoc_annotations',
        'autodoc:class': 'autodoc_class',
        'autodoc:module': 'autodoc_module',
        'c:last_symbol': 'c_last_symbol',
        'c:namespace_stack': 'c_namespace_stack',
        'c:parent_symbol': 'c_parent_symbol',
        'cpp:domain_name': 'cpp_domain_name',
        'cpp:last_symbol': 'cpp_last_symbol',
        'cpp:namespace_stack': 'cpp_namespace_stack',
        'cpp:parent_symbol': 'cpp_parent_symbol',
        'default_domain': 'default_domain',
        'default_role': 'default_role',
        'docname': 'docname',
        'highlight_language': 'highlight_language',
        'object': 'obj_desc_name',
        'started_at': 'reading_started_at',
    }

    # Attributes that should reset to None if popped.
    __attr_default_none: Final = frozenset({
        '_parser',
        'c:last_symbol',
        'c:parent_symbol',
        'cpp:last_symbol',
        'cpp:parent_symbol',
        'default_domain',
    })

    def __init__(
        self,
        *,
        docname: str = '',
        default_role: str = '',
        default_domain: Domain | None = None,
    ) -> None:
        #: The docname of the document currently being parsed.
        self.docname: str = docname

        #: The default role for the current document.
        #: Set by the ``.. default-role::`` directive.
        self.default_role: str = default_role

        #: The default domain for the current document.
        #: Set by the ``.. default-domain::`` directive.
        self.default_domain: Domain | None = default_domain

        #: The parser being used to parse the current document.
        self._parser: Parser | None = None

        #: The default language for syntax highlighting.
        #: Set by the ``.. highlight::`` directive to override
        #: the ``highlight_language`` config value.
        self.highlight_language: str = ''

        #: The current object's name.
        #: Used in the Changes builder.
        self.obj_desc_name: str = ''

        #: Records type hints of Python objects in the current document.
        #: Used in ``sphinx.ext.autodoc.typehints``.
        #: Maps object names to maps of attribute names -> type hints.
        self.autodoc_annotations: dict[str, dict[str, str]] = {}

        #: The current Python class name.
        #: Used in ``sphinx.ext.autodoc``.
        self.autodoc_class: str = ''

        #: The current Python module name.
        #: Used in ``sphinx.ext.autodoc``.
        self.autodoc_module: str = ''

        #: The most-recently added declaration in a directive.
        #: Used in the C Domain.
        self.c_last_symbol: CSymbol | None = None

        #: The stack of namespace scopes, altered by the ``.. c:namespace::``
        #: and ``.. c:namespace-(push|pop)::``directives.
        #: Used in the C Domain.
        self.c_namespace_stack: list[CSymbol] = []

        #: The parent declaration.
        #: Used in the C Domain.
        self.c_parent_symbol: CSymbol | None = None

        #: A stack of the string representation of declarations,
        #: used to format the table of contents entry.
        #: Used in the C++ Domain.
        self.cpp_domain_name: tuple[str, ...] = ()

        #: The most-recently added declaration in a directive.
        #: Used in the C++ Domain.
        self.cpp_last_symbol: CPPSymbol | None = None

        #: The stack of namespace scopes, altered by the ``.. cpp:namespace::``
        #: and ``.. cpp:namespace-(push|pop)::``directives.
        #: Used in the C++ Domain.
        self.cpp_namespace_stack: list[CPPSymbol] = []

        #: The parent declaration.
        #: Used in the C++ Domain.
        self.cpp_parent_symbol: CPPSymbol | None = None

        #: Records the time when reading begain for the current document.
        #: Used in ``sphinx.ext.duration``.
        self.reading_started_at: float = 0.0

        # Used for generating unique serial numbers.
        self._serial_numbers: dict[str, int] = {}

        # Stores properties relating to the current document set by extensions.
        self._extension_data: dict[str, Any] = {}

    def new_serial_number(self, category: str = '', /) -> int:
        """Return a serial number, e.g. for index entry targets.

        The number is guaranteed to be unique in the current document & category.
        """
        current = self._serial_numbers.get(category, 0)
        self._serial_numbers[category] = current + 1
        return current

    # Mapping interface:

    def __getitem__(self, item: str) -> Any:
        if item in self.__attr_map:
            return getattr(self, self.__attr_map[item])
        return self._extension_data[item]

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self.__attr_map:
            setattr(self, self.__attr_map[key], value)
        else:
            self._extension_data[key] = value

    def __delitem__(self, key: str) -> None:
        self.pop(key, default=None)

    def __contains__(self, item: str) -> bool:
        if item in {'c:parent_symbol', 'cpp:parent_symbol'}:
            return getattr(self, self.__attr_map[item]) is not None
        return item in self.__attr_map or item in self._extension_data

    def __iter__(self) -> Iterator[str]:
        return iter(self.keys())

    def __len__(self) -> int:
        return len(self.__attr_map) + len(self._extension_data)

    def keys(self) -> Iterable[str]:
        return frozenset(self.__attr_map.keys() | self._extension_data.keys())

    def items(self) -> Iterable[tuple[str, Any]]:
        for key in self.keys():
            yield key, self[key]

    def values(self) -> Iterable[Any]:
        for key in self.keys():
            yield self[key]

    def get(self, key: str, default: Any | None = None) -> Any | None:
        try:
            return self[key]
        except KeyError:
            return default

    __sentinel = object()

    def pop(self, key: str, default: Any | None = __sentinel) -> Any | None:
        if key in self.__attr_map:
            # the keys in  __attr_map always exist, so ``default`` is ignored
            value = getattr(self, self.__attr_map[key])
            if key in self.__attr_default_none:
                default = None
            else:
                default = type(value)()  # set key to type's default
            setattr(self, self.__attr_map[key], default)
            return value
        if default is self.__sentinel:
            return self._extension_data.pop(key)
        return self._extension_data.pop(key, default)

    def setdefault(self, key: str, default: Any | None = None) -> Any | None:
        return self._extension_data.setdefault(key, default)

    def clear(self) -> None:
        _CurrentDocument.__init__(self)  # NoQA: PLC2801

    def update(self, other: Iterable[tuple[str, Any]] = (), /, **kwargs: Any) -> None:
        other_dict = dict(other) if not isinstance(other, dict) else other
        for dct in other_dict, kwargs:
            for key, value in dct.items():
                self[key] = value
